import pandas as pd
import os, re
import getpass
import kblab
import progressbar
from lxml import etree
from .utils import clean_html
import warnings
import alto

class LazyArchive:
    """
    Wrapper to KBLab archive class so that you don't need to
    log in if you don't actually use the archive
    """

    def __init__(self):
        self.archive = None

    def __getattr__(self, attr):
        if self.archive == None:
            self.archive = _login_to_archive()
        return getattr(self.archive, attr)


def _login_to_archive():

    if "KBLPASS" in os.environ and "KBLUSER" in os.environ and "KBLMYLAB" in os.environ:
        print("Connecting to KB the laziest way...")
        return kblab.Archive(os.environ.get("KBLMYLAB"), auth=(os.environ.get("KBLUSER"), os.environ.get("KBLPASS")))

    else:
        username = input("Username: ")
        password = getpass.getpass()
        print("Password set for user:", username)
        return kblab.Archive("https://betalab.kb.se", auth=(username, password))


def oppna_data_to_dict(input_dict):
    """
    Load protocols with the new XML / HTML structure (from 2013 onwards)
    and convert it to a python dict with contents.
    """
    data = {}
    data["paragraphs"] = []

    # Metadata
    session = input_dict["dokumentstatus"]["dokument"]["rm"]
    session = session.replace("/", "")
    pid = input_dict["dokumentstatus"]["dokument"]["nummer"]
    date = input_dict["dokumentstatus"]["dokument"]["datum"]
    html = input_dict["dokumentstatus"]["dokument"]["html"]
    html_tree = clean_html(html)
    year = int(date.split("-")[0])
    protocol_id = f"prot-{session}--{pid}"

    data["protocol_id"] = protocol_id
    data["date"] = date.split(" ")[0]
    data["session"] = session

    # New HTML structure with div[@class='Section1']
    section1 = html_tree.xpath(".//div[@class='Section1']")
    for elements in section1:
        for elem in elements:
            if elem.tag in ["p", "h1", "h2"]:
                elemtext = "".join(elem.itertext())
                linebreak = elemtext.strip() == "" and "\n" in elemtext
                if linebreak:
                    pass
                else:
                    paragraph = elemtext.strip()
                    paragraph = paragraph.replace("\n", " ")
                    paragraph = re.sub("\\s+", " ", paragraph)
                    data["paragraphs"].append(paragraph)

    if len(data["paragraphs"]) == 0:
        tree = html_tree

        # Old data structure 1990-2003
        pres = tree.findall(".//pre")
        if len(pres) > 0:
            for pre in pres:
                if pre.text is not None:
                    tblocks = re.sub("([a-zß-ÿ,])- ?\n ?([a-zß-ÿ])", "\\1\\2", pre.text)
                    tblocks = re.sub("([a-zß-ÿ,]) ?\n ?([a-zß-ÿ])", "\\1 \\2", tblocks)
                    for paragraph in tblocks.split("\n"):
                        paragraph = paragraph.replace("\n", " ")
                        paragraph = paragraph.replace("\n", " ")
                        data["paragraphs"].append(paragraph)

        # Standard HTML structure, roughly 2003-2013
        elif len(tree.xpath("//div[@class='indrag']")) > 0:
            tree = tree.xpath("//body")[0]
            for elem in tree:
                elemtext = "".join(elem.itertext())
                linebreak = elemtext.strip() == "" and "\n" in elemtext
                if elem.tag == "br" or linebreak:
                    pass
                else:
                    paragraph = elemtext.strip()
                    paragraph = paragraph.replace("\n", " ")
                    paragraph = re.sub("\\s+", " ", paragraph)
                    data["paragraphs"].append(paragraph)
    return data

def _alto_extract_paragraphs(altofile):
    """
    Extract text from ALTO XML on paragraph / textBlock level
    """
    paragraphs = []
    text_blocks = altofile.extract_text_blocks()
    for tb_ix, tb in enumerate(text_blocks):
        lines = tb.extract_string_lines()            
        paragraph = "\n".join(lines)
        
        # Remove line breaks when next line starts with a small letter
        paragraph = re.sub("([a-zß-ÿ,])- ?\n ?([a-zß-ÿ])", "\\1\\2", paragraph)
        paragraph = re.sub("([a-zß-ÿ,]) ?\n ?([a-zß-ÿ])", "\\1 \\2", paragraph)
        
        paragraph = " ".join(paragraph.split())
        if paragraph != "":
            paragraphs.append(paragraph)
    return paragraphs

def convert_alto(filenames, files):
    """
    Convert a document from ALTO to a list of paragraphs.

    Args:
        filenames: the names of the ALTO files of one document, as a list of str.
            The script assumes zero-padded numbering right before the .xml extension.
        files: ALTO XML files as a list of str in corresponding order to the filenames
    """
    in_sync = True
    paragraphs = []
    for ix, pair in progressbar.progressbar(enumerate(zip(filenames, files))):
        fname, s = pair
        altofile = alto.parse(s)
        page_number = int(re.findall("([0-9]{3,3}).xml", fname)[0])
        paragraphs.append(page_number)
        if in_sync and page_number != ix:
            not_in_sync_warning = f"ALTO page number and page count not in sync ({fname})"
            warnings.warn(not_in_sync_warning)
            in_sync = False
        paragraphs += _alto_extract_paragraphs(altofile)
    return paragraphs

def dl_kb_blocks(package_id, archive):
    """
    Download protocol from betalab, convert it to the simple XML 'blocks' schema
    """
    print(f"Get package {package_id}...")
    package = archive.get(package_id)
    filenames = fetch_files(package)
    def files():
        for fname in filenames:
            yield package.get_raw(fname).read()

    return convert_alto(filenames, files())


def count_pages(start, end):
    """
    Generate a dataframe of pages between provided start and end years. Fetches information from KB's API.
    """
    years = range(start, end)
    archive = LazyArchive()
    rows = []

    for year in progressbar.progressbar(years):
        params = {"tags": "protokoll", "meta.created": str(year)}
        package_ids = archive.search(params, max=365)

        for package_id in package_ids:
            package = archive.get(package_id)
            jp2list = fetch_files(package, extension="jp2")
            page_count = len(jp2list)
            rows.append([package_id, year, page_count])

    columns = ["protocol_id", "year", "pages"]
    db_pages = pd.DataFrame(rows, columns=columns)
    return db_pages


def _create_dirs(outfolder):
    if not os.path.exists(outfolder):
        print("Create folder", outfolder)
        os.mkdir(outfolder)

    if not os.path.exists(outfolder + "train/"):
        os.mkdir(outfolder + "train/")

    if not os.path.exists(outfolder + "test/"):
        os.mkdir(outfolder + "test/")


def fetch_files(package, extension="xml"):
    """
    Fetch all files with the provided extension from a KBLab package

    Args:
        package: KBLab client package
        extension: File extension of the files that you want to fetch.
        String, or None which outputs all filetypes.
    """
    filelist = package.list()
    if extension is not None:
        filelist = [f for f in filelist if f.split(".")[-1] == extension]
    filelist = sorted(filelist)

    return filelist


def _get_seed(string):
    encoded = string.encode("utf-8")
    digest = hashlib.md5(encoded).hexdigest()[:8]
    return int(digest, 16)


def randomize_ordinals(files):
    """
    Create pseudo-random ordinal numbers for a database
    """
    columns = ["package_id", "year", "pagenumber", "ordinal"]
    data = []
    for index, row in files.iterrows():
        # print(index, row)
        package_id = row["package_id"]
        pages = row["pages"]
        year = row["year"]

        for page in range(0, pages):

            seedstr = package_id + str(year) + str(page)
            np.random.seed(_get_seed(seedstr))
            ordinal = np.random.rand()
            new_row = [package_id, year, page, ordinal]
            data.append(new_row)

    return pd.DataFrame(data, columns=columns)



from SPARQLWrapper import SPARQLWrapper, JSON
import numpy as np
import pandas as pd
from importlib_resources import files

def impute_query_string(source=None):
	# Defaults to all individuals
	if not source:
		source = ["member_of_parliament", "minister", "speaker"]
	
	strings = []
	for src in source:
		# Specific individual wikidata object
		if src.startswith('Q'):
			s = f"VALUES ?wiki_id {{ wd:{src} }}"
		# Category of wikidata objects
		else:
			s = files('pyriksdagen.data.queries').joinpath(f'{src}.txt').read_text()
		strings.append(s)
	return "\n} UNION {\n".join(strings)

def get_query_string(query_name):
	try:
		s = files('pyriksdagen.data.queries').joinpath(f'{query_name}.rq').read_text()
	except FileNotFoundError:
		print(f"File {query_name} not found")
	return s

def reduce_date_precision(date, precision):
	'''
	Truncates date to accurate length given precision level
	'''
	# Decade
	if int(precision) == 8:
		return str(int(date[:4]) // 10 * 10)
	# Year
	if int(precision) == 9:
		return date[:4]
	# Month
	if int(precision) == 10:
		return date[:7]
	# Day
	if int(precision) == 11:
		return date[:10]

def fix_dates(df):
	'''
	Converts date precision for all relevant cols and removes the precision columns
	'''
	for c in df.columns:
		if 'Precision' in c:
			base_name = c.replace('Precision', '')
			df.loc[df[base_name].notna(), base_name] =\
			df.loc[df[base_name].notna()].apply(lambda x: reduce_date_precision(x[base_name], x[c]), axis=1)
			df = df.drop(c, axis=1)
	return df

def clean_sparql_df(df, query_name):
	df = df.copy()

	# Clean columns
	df = df.rename(columns={'wiki_idLabel.value':'name'}) # avoid duplicate colnames
	df = df.rename(columns={'government.value':'government_id.value'})
	df = df.rename(columns={'party.value':'party_id.value'})
	df = df[[c for c in df.columns if c.endswith('.value') or c == 'name']]
	
	df.columns = df.columns.str.replace('.value', '', regex=False)
	df.columns = df.columns.str.replace('Label', '', regex=False)
	df = fix_dates(df) # use and drop date precision columns

	# Format values
	for colname in df.columns:
		if "_id" in colname:
			df[colname] = df[colname].str.split('/').str[-1]

	# Drop pseudo missing values of form "http://www.wikidata.org/.well-known..."
	idx, idy = np.where(df.astype(str).applymap(lambda x: 'http' in x))
	for x, y in zip(idx, idy):
		df.loc[x][y] = ''

	# Sort columns
	first_cols = [c for c in ['person_id', 'wiki_id', 'start', 'end'] if c in df.columns]
	other_cols = sorted([c for c in df.columns if c not in first_cols])
	df = df[first_cols+other_cols]

	# Sort rows
	first_cols.reverse()
	df = df.sort_values(by=list(first_cols+other_cols))
	return df

def query2df(query_name, source=None):
	query = get_query_string(query_name)
	source = impute_query_string(source)
	query = query.replace('PLACEHOLDER', source)
	sparql = SPARQLWrapper("https://query.wikidata.org/sparql", agent="Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.11 (KHTML, like Gecko) Chrome/23.0.1271.64 Safari/537.11")
	sparql.setQuery(query)	
	sparql.setReturnFormat(JSON)
	try:
		results = sparql.query().convert()
	except:
		print(f"Query {query_name} failed")
	df = pd.json_normalize(results['results']['bindings'])
	df = clean_sparql_df(df, query_name)
	return df

def separate_name_location(name_location_specifier, alias):
	alias = alias[['person_id', 'alias']].rename(columns={'alias':'name'})
	primary_df = name_location_specifier[['person_id', 'name']]
	secondary_df = name_location_specifier[['person_id', 'alias']].rename(columns={'alias':'name'})
	primary_df['primary_name'] = True
	secondary_df['primary_name'] = False
	alias['primary_name'] = False
	df = pd.concat([primary_df, secondary_df, alias]).dropna().drop_duplicates().reset_index(drop=True)

	# Split names and location specifiers
	names = df['name'].str.split(' [io] ', expand=True)
	loc_cols = [i for i in range(len(df.columns)-1)]
	names.columns = ['name']+loc_cols
	df = df.drop('name', axis=1)
	df = df.join(names, how='left')

	# Cleaning
	df = df[~df['name'].str.contains(',')]
	df[loc_cols] = df[loc_cols].apply(lambda x: x.str.replace('och','').str.strip())
	df['name'] = df['name'].str.replace(r'\([^()]*\)', '', regex=True)
	df['name'] = df['name'].apply(lambda x: ' '.join(x.split()))

	# Drop duplicates
	name = 	df[['person_id', 'name', 'primary_name']].\
			sort_values(by=['primary_name'], ascending=False).\
			drop_duplicates(subset=['person_id', 'name', 'primary_name'])

	loc =	pd.concat([df[['person_id', col]].rename(columns={col:'location'}) for col in loc_cols]).\
			dropna().drop_duplicates()

	# Sort values
	name = name[['person_id'] + sorted([col for col in name.columns if col != 'person_id'])]
	name = name.sort_values(by=list(name.columns))
	loc = loc[['person_id'] + sorted([col for col in loc.columns if col != 'person_id'])]
	loc = loc.sort_values(by=list(loc.columns))
	return name, loc

def move_party_to_party_df(mp_df, party_df):
	mp_parties = mp_df[party_df.columns]
	mp_parties = mp_parties[mp_parties['party_id'].notnull()]

	mp_parties = mp_parties.sort_values(["person_id", "start"])
	party_df = pd.concat([mp_parties, party_df])
	party_df = party_df.drop_duplicates()

	mp_df_cols = [col for col in mp_df.columns if col not in ["party", "party_id"]]

	return mp_df[mp_df_cols], party_df

def elongate_external_ids(df):
	rows = []
	cols = ["person_id", "authority", "identifier"]
	df.rename(columns={"wiki_id": "WiDaID"}, inplace=True)
	authorities = [_ for _ in df.columns if _ != "person_id"]
	for i, r in df.iterrows():
		swerik = r["person_id"]
		for a in authorities:
			if pd.notnull(r[a]):
				rows.append([swerik, a, r[a]])
	return pd.DataFrame(rows, columns=cols)



"""
Implements the segmentation of the data into speeches and
ultimately into the Parla-Clarin XML format.
"""
import numpy as np
import re, hashlib
from .db import load_expressions
from .match_mp import match_mp, name_equals, name_almost_equals, names_in, names_in_rev
from itertools import combinations
from trainerlog import get_logger
LOGGER = get_logger("segmentation")

# Classify paragraph
def classify_paragraph(paragraph, classifier, prior=np.log([0.8, 0.2])):
    """
    Classify paragraph into speeches / descriptions with provided classifier
    
    Args:
        paragraph (str): the text content of a paragraph
        classifier (dict): a dictionary that includes 'dim', 'ft' and 'model' keys for classification
        prior (np.array): log prior for the classes
    """
    words = paragraph.split()
    V = len(words)
    if V == 0:
        return prior

    x = np.zeros((V, classifier["dim"]))

    ft = classifier["ft"]
    for ix, word in enumerate(words):
        vec = ft.get_word_vector(word)
        x[ix] = vec

    pred = classifier["model"].predict(x, batch_size=V)
    return np.sum(pred, axis=0) + prior


def detect_speaker(matched_txt, speaker_db, metadata=None):
    """
    Detect the speaker of the house

    Args:
        matched_txt (str): intro text
        speaker_db (pd.df): dataframe containing the speaker metadata
        metadata (dict): metadata about the protocol. Deprecated.

    Returns
        speaker_id (str): ID as a string if detected, otherwise None
    """
    lower_txt = matched_txt.lower()

    # Second vice speaker
    if re.search('andre vice', lower_txt):
        speaker_db = speaker_db[speaker_db["role"].str.contains('andre')]
        
    # Third vice speaker
    elif re.search('tredje vice', lower_txt):
        speaker_db = speaker_db[speaker_db["role"].str.contains('tredje')]

    # First vice speaker
    elif re.search(r'(förste)?\svice', lower_txt):
        speaker_db = speaker_db[speaker_db["role"].str.contains('förste')]

    # Speaker
    elif re.search(r'(herr|fru)?\s?talman', lower_txt):
        speaker_db = speaker_db[speaker_db["role"] == 'talman']

    number_of_matches = len(set(speaker_db["id"]))
    if number_of_matches == 1:
        matched_value = speaker_db["id"].iloc[0]
        LOGGER.debug(f"Match found: {matched_value}")
        return matched_value
    elif number_of_matches >= 2:
        matched_values = set(speaker_db["id"])
        LOGGER.debug(f"Ambiguous matches: {matched_values}")
    else:
        LOGGER.debug(f"No matches")


def detect_minister(matched_txt, minister_db, intro_dict):
    """
    Detect a minister

    Args:
        matched_txt (str): intro text
        minister_db (pd.df): dataframe containing the minister metadata
        intro_dict (dict): processed information about the intro text, possibly containing eg. 'gender' 
    
    Returns:
        ministed_id (str): ID as a string if detected, otherwise None
    """
    lower_txt = matched_txt.lower()

    # Filter by gender
    if 'gender' in intro_dict:
        gender = intro_dict["gender"]
        minister_db = minister_db[minister_db["gender"] == gender]

    # Filter by date
    if 'date' in intro_dict:

        minister_db = minister_db[
                (minister_db["start"] <= intro_dict['date']) &
                (minister_db["end"] >= intro_dict['date'])]
        if not minister_db.empty:
            if len(set(minister_db["id"])) == 1:
                return minister_db["id"].iloc[0]

    # Match by name
    if 'name' in intro_dict:
        name = intro_dict["name"].lower()
        # thage petterson
        #print(minister_db)
        name_matches = names_in(name, minister_db)
        if not name_matches.empty:
            if len(set(name_matches["id"])) == 1:
                matched_value = name_matches["id"].iloc[0]
                LOGGER.debug(f"Match by name {matched_value}")
                return matched_value

    # Match by role
    # Catch "utrikesdepartementet"
    if role := re.search(r'([A-Za-zÀ-ÿ]+)(?:departementet)', lower_txt):
        r = role.group(0).replace('departementet', '')
        role_matches = minister_db[minister_db["role"].str.contains(r, regex=False)]
        if not role_matches.empty:
            if len(set(role_matches["id"])) == 1:
                matched_value = role_matches["id"].iloc[0]
                LOGGER.debug(f"Matched by role {matched_value}")
                return matched_value

    # Catch "ministern för utrikes ärendena (...)"
    elif role := re.search(r'(?:ministern för )([A-Za-zÀ-ÿ]+)', lower_txt):
        r = role.group(0).split()[-1]
        role_matches = minister_db[minister_db["role"].str.contains(r, regex=False)]
        if not role_matches.empty:
            if len(set(role_matches["id"])) == 1:
                return role_matches["id"].iloc[0]

    elif role := re.search(r'[A-Za-zÀ-ÿ]+minister', lower_txt):
        r = role.group(0).replace('minister', '')
        role_matches = minister_db[minister_db["role"].str.contains(r, regex=False)]
        if not role_matches.empty:
            if len(set(role_matches["id"])) == 1:
                return role_matches["id"].iloc[0]

def detect_mp(intro_dict, db, party_map=None, match_fuzzily=False):
    """
    Match an MP in a text snippet. Returns an MP id (str) if found, otherwise None.

    If multiple people are matched, defaults to returning None.
    """
    LOGGER.debug(f"Detect MP based on: {intro_dict}")

    intro_dict["party_abbrev"] = party_map.get(intro_dict.get("party", ""), "")
    variables = ['party_abbrev', 'specifier', 'name']
    variables = [v for v in variables if v in list(db.columns)] # removes missing variables
    variables = sum([list(map(list, combinations(variables, i))) for i in range(len(variables) + 1)], [])[1:]
    if match_fuzzily:
        matching_funs = [name_equals, name_almost_equals, names_in]
    else:
        matching_funs = [name_equals, names_in]
    return match_mp(intro_dict, db, variables, matching_funs)

def intro_to_dict(intro_text, expressions=None):
    """
    Convert introduction to a metadata dictionary.
    Tries to detect 'name', 'gender', 'party', 'specified', as well as some misc things labeled as 'other'.

    Args:
        intro_text (str): introduction text
        expressions (list): a list of regular expressions used for metadata detection.
            Used for optimization, the algorithm works the same if not provided.
    
    Returns:
        d (dict): detected metadata
    """
    if expressions is None:
        expressions = load_expressions(phase="mp")
    intro_text = intro_text.strip()
    d = {}
    for exp, t in expressions:
        m = exp.search(intro_text)
        if m is not None:
            matched_text = m.group(0)

            if t not in d:
                d[t] = matched_text.strip()
                intro_text = intro_text.replace(matched_text, " ")

    if "name" in d:
        s = d["name"]
        s = s.replace('-', ' ')
        if ", " in s:
            s = s.split(", ")
            s = s[1] + " " + s[0]
        d["name"] = s

    if "gender" in d:
        d["gender"] = d["gender"].lower()
        if d["gender"] in ["herr", "friherre"]:
            d["gender"] = "man"
        if d["gender"] in ["fru", "fröken"]:
            d["gender"] = "woman"

    if "specifier" in d:
        d["specifier"] = d["specifier"].replace("i ", "")
        d["specifier"] = d["specifier"].replace("från ", "")

    return d

def expression_dicts(pattern_db):
    expressions = dict()
    manual = dict()
    for _, row in pattern_db.iterrows():
        pattern = row["pattern"]
        exp = re.compile(pattern)
        # Calculate digest for distringuishing patterns without ugly characters
        pattern_digest = hashlib.md5(pattern.encode("utf-8")).hexdigest()[:16]
        expressions[pattern_digest] = exp
    return expressions, manual


def detect_introduction(elem, intro_ids):
    """
    Detect whether the current paragraph contains an introduction of a speaker.

    Returns a dict if an intro is detected, otherwise None.
    """
    if elem.attrib.get("{http://www.w3.org/XML/1998/namespace}id") in intro_ids:

            d = {
                "pattern": None,
                "who": None,
                "segmentation": None,
                "txt": elem.text,
            }

            return d

def combine_intros(elem1, elem2, intro_expressions, other_expressions):
    """
    Join intros that have been split as an artifact of the data processing.
    """
    if elem1.text is None or elem2.text is None:
        return False
    combine = False
    for exp, _ in other_expressions:
        for m in exp.finditer(elem1.text.strip()):
            combine = True

    intro = detect_introduction(elem2.text, intro_expressions)
    combine = combine and intro is not None
    combine = combine and "Anf" not in elem2.text
    if combine:
        if elem1.text.strip()[-1] == "-":
            elem2.text = elem1.text.strip()[:-1] + "-" + elem2.text.strip()
        else:
            elem2.text = elem1.text + " " + elem2.text
        elem1.text = ""

    return combine

def join_text(text1, text2):
    text1, text2 = list(map(lambda x: ' '.join(x.replace('\n', ' ').split()), [text1, text2]))
    # Account for words split over textblocks with '-'
    if text1.endswith('-'):
        return ''.join([text1[:-1], text2])
    else:
        return ' '.join([text1, text2])



"""
Handles the data on the members of parliament.
"""

import pandas as pd
import os, re
import progressbar
import hashlib
import unicodedata


def create_database(path):
    """
    Create an initial version of the MP dataframe
    """
    extension = path.split(".")[-1]

    if extension == "csv":
        print("Read:", path)
        df = pd.read_csv(path, skip_blank_lines=True)

        # Drop columns where everything is null
        nulls = df.isnull().values.all(axis=0)
        nulls = zip(df.columns, nulls)
        for column_name, null in nulls:
            if null:
                del df[column_name]

        new_columns = list(df.columns)
        for column_ix, column_name in enumerate(df.columns):
            if "." in column_name:
                new_name = column_name.split(".")[0]
                new_columns[column_ix] = new_name

        df.columns = new_columns

    elif extension == "txt":
        print("Read:", path)
        f = open(path)
        columns = ["Riksdagsledamot", "Parti", "Valkrets"]

        rows = []
        lan = None
        for line in f:
            line = line.replace("\n", "")
            if len(line) > 2:
                indented = line[:4] == "    "

                # Non-indented lines are titles, which correspond to 'län' / region
                if not indented:
                    lan = line
                else:
                    row = line.split(",")
                    row = [x.strip() for x in row]

                    datapoint = []
                    # Add name
                    name = row[0]
                    datapoint.append(name)
                    # Add party
                    possible_parties = [s for s in row if "f." not in s]
                    party = possible_parties[-1]
                    party = re.sub(r"\(.*?\)", "", party)
                    party = re.sub(r"\[.*?\]", "", party).strip()
                    if len(party) < 4 or party.lower() != party:
                        datapoint.append(party)
                    else:
                        datapoint.append(None)
                    # Add 'län' / region
                    datapoint.append(lan)
                    rows.append(datapoint)

        df = pd.DataFrame(rows, columns=columns)
    else:
        print("File type not supported.", path)
        return None

    # Harmonize column names
    new_columns = list(df.columns)
    for column_ix, column_name in enumerate(df.columns):
        if column_name in ["Ledamot", "Riksdagsledamot", "Namn"]:
            new_columns[column_ix] = "name"
        if column_name in ["Parti"]:
            new_columns[column_ix] = "party"
        if column_name in ["Yrke"]:
            new_columns[column_ix] = "occupation"
        if column_name in ["Valkrets"]:
            new_columns[column_ix] = "district"

    df.columns = new_columns

    # Drop unnecessary columns
    retain = ["name", "party", "district", "occupation"]
    for column_name in df.columns:
        if column_name not in retain:
            del df[column_name]

    # Chamber
    chamber = "Enkammarriksdagen"
    potential_chamber = path.split("/")[-2]
    if potential_chamber == "ak":
        chamber = "Andra kammaren"
    elif potential_chamber == "fk":
        chamber = "Första kammaren"
    df["chamber"] = chamber

    # Year in office
    year_str = path.split("/")[-1].split(".")[0].replace("–­­", "-")
    if len(year_str) == 9:
        df["start"] = int(year_str[:4])
        df["end"] = int(year_str[-4:])
    elif len(year_str) == 4:
        # TODO: Currently using +-3 year heuristic because exact terms
        # are unavailable.
        if chamber == "Första kammaren":
            df["start"] = int(year_str) - 3
            df["end"] = int(year_str) + 3
        else:
            df["start"] = int(year_str)
            df["end"] = int(year_str)
    else:
        print(year_str)

    return df


def create_full_database(dirs):
    mp_dbs = []
    for d in dirs:
        for path in os.listdir(d):
            full_path = os.path.join(d, path)
            mp_db = create_database(full_path)
            if mp_db is not None:
                mp_dbs.append(mp_db)
    mp_db = pd.concat(mp_dbs)

    mp_db = mp_db.sort_values(by=["start", "chamber", "name"], ignore_index=True)

    columnsTitles = [
        "name",
        "party",
        "district",
        "chamber",
        "start",
        "end",
        "occupation",
    ]
    mp_db = mp_db.reindex(columns=columnsTitles)

    mp_db = mp_db[mp_db["name"].notnull()]

    print(mp_db[mp_db["start"].isnull()])
    mp_db.start = mp_db.start.astype(int)
    mp_db.end = mp_db.end.astype(int)
    return mp_db


def add_gender(mp_db, names):
    """
    Based to first names, add gender to an MP dataframe.
    """
    print("Add gender...")
    mp_db["gender"] = None

    name_to_gender = {}
    for i, namerow in names.iterrows():
        name = namerow["name"]
        gender = namerow["gender"]
        if gender == "masculine":
            gender = "man"
        elif gender == "feminine":
            gender = "woman"
        name_to_gender[name] = gender

    for i, row in progressbar.progressbar(list(mp_db.iterrows())):
        first_name = row["name"].split()[0]
        if "-" in first_name:
            first_name = first_name.split("-")[0]
        if first_name in name_to_gender:
            mp_db.loc[i, "gender"] = name_to_gender[first_name]

    return mp_db


def clean_names(mp_db):
    """
    Remove artefacts from MP names and specifiers
    """
    print("Clean names...")
    for i, row in progressbar.progressbar(list(mp_db.iterrows())):
        name = row["name"]
        split_i = name.split(" i ")
        if name != split_i[0]:
            name = split_i[0]
            if len(split_i) > 1:
                mp_db.loc[i, "specifier"] = "i " + split_i[1]
            else:
                mp_db.loc[i, "specifier"] = None
        if "[" in name:
            name = name.split("[")[0]
        if "(er" in name:
            name = name.split("(er")[0]
        if "ersatt av" in name:
            name = name.split("ersatt av:")[-1]
        name = name.strip()
        assert name != "", "names can't be empty: " + row["name"]
        mp_db.loc[i, "name"] = name

    return mp_db


def replace_party_abbreviations(mp_db, party_db):
    """
    Replace party abbreviations with standardized party names.
    """
    print("Replace party abbreviations...")
    party_dict = dict()
    for _, row in party_db.iterrows():
        party = row["party"]
        abbreviation = row["abbreviation"]
        party_dict[abbreviation] = party

    for i, row in progressbar.progressbar(list(mp_db.iterrows())):
        current_party = row["party"]
        if type(current_party) == str:
            # Check if 'party' attribute is in the list of abbreviations
            row_party = current_party.strip().lower()
            if row_party in party_dict:
                mp_db.loc[i, "party"] = party_dict[row_party]

    return mp_db


def add_id(mp_db):
    """
    Generate deterministic IDs for mps based on the "name", "party", "district",
    "chamber", "start", and "end" columns of the dataframe.
    """
    print("Add id...")
    columns = mp_db.columns
    columns = ["name", "party", "district", "chamber", "start", "end"]
    mp_db["id"] = None
    print("columns used for generation:", ", ".join(columns))
    for i, row in progressbar.progressbar(list(mp_db.iterrows())):
        name = unicodedata.normalize("NFD", row["name"])
        name = name.encode("ascii", "ignore").decode("utf-8")
        name = name.lower().replace(" ", "_")
        name = name.replace(".", "").replace("(", "").replace(")", "").replace(":", "")
        party = row.get("party")

        pattern = [name]
        for column in columns:
            value = row[column]
            if type(value) != str:
                value = str(value)
            pattern.append(value)

        pattern = "_".join(pattern).replace(" ", "_").lower()

        digest = hashlib.md5(pattern.encode("utf-8")).hexdigest()
        mp_db.loc[i, "id"] = name + "_" + digest[:6]

    return mp_db


def add_municipality(mp_db, mun_db):
    """
    Add home municipalities as specifiers for matched MPs from personregister
    """
    original_columns = list(mp_db.columns)
    print("Add municipalicites from 'personregister'...")

    def reorder_name(name):
        s = name.split(",")
        if len(s) == 1:
            return name
        else:
            newname = s[1].strip() + " " + s[0].strip()
            return newname

    mun_db["name"] = mun_db["name"].apply(lambda n: reorder_name(n))
    mun_db = mun_db[mun_db["municipality"].notnull()]
    mun_db["municipality"].apply(lambda x: "i " + x.strip())
    mun_db["decade"] = mun_db["decade"].apply(lambda decade: (decade // 10) * 10)

    outdfs = []
    # mp_db["municipality"] = None

    start_min = min(set(mp_db["start"]))
    end_max = max(set(mp_db["end"]))

    start_min = (start_min // 10) * 10
    end_max = (end_max // 10 + 1) * 10
    for decade in range(start_min, end_max, 10):
        current_mun_db = mun_db[mun_db["decade"] == decade]
        current_mpdb = mp_db[(mp_db["end"] >= decade) & (mp_db["start"] < decade + 10)]
        if not current_mun_db.empty:
            current_mun_db = current_mun_db[["name", "municipality"]]

            mpdb_names = list(current_mpdb["name"])

            def rname(name):
                nameset = set(name.split())

                namesplit = name.split()
                for mpdb_name in mpdb_names:
                    mpdbsplit = mpdb_name.split()
                    samefirst = namesplit[0] == mpdbsplit[0]
                    samelast = namesplit[-1] == mpdbsplit[-1]
                    if (
                        samefirst
                        and samelast
                        and abs(len(namesplit) - len(mpdbsplit)) <= 1
                    ):
                        return mpdb_name.strip()
                return name.strip()

            replacedname = rname("Carl Wilhelm Höglund")

            print("Carl Wilhelm Höglund", "=>", replacedname, decade)
            current_mun_db["name"] = current_mun_db["name"].apply(lambda n: rname(n))
            merged = pd.merge(current_mpdb, current_mun_db, how="left", on="name")
            merged.to_csv("merged_" + str(decade) + ".csv", index=False)

            newnames = set(current_mun_db["name"])
            outdfs.append(merged)
            print(
                "Carl Wilhelm Oskar Höglund in newnames",
                "Carl Wilhelm Oskar Höglund" in newnames,
            )
        else:
            outdfs.append(current_mpdb)

    mp_db = pd.concat(outdfs)
    print(mp_db)
    mp_db["municipality"] = "i " + mp_db["municipality"].str.strip()
    mp_db["specifier"] = mp_db["specifier"].fillna(mp_db["municipality"])
    mp_db = mp_db.drop("municipality", axis=1)

    mp_db = mp_db.groupby("id").first().reset_index()
    mp_db = mp_db[original_columns]
    mp_db = mp_db.sort_values(by=["start", "chamber", "name"], ignore_index=True)
    return mp_db



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Provides useful utilities for the other modules as well as for general use.
"""
from bs4 import BeautifulSoup
from datetime import datetime
from lxml import etree
from pathlib import Path
from pyparlaclarin.refine import format_texts
from tqdm import tqdm
from trainerlog import get_logger
import base58
import hashlib
import os
import requests
import uuid
import warnings
import xmlschema
import zipfile




LOGGER = get_logger("pyriksdagen")
XML_NS = "{http://www.w3.org/XML/1998/namespace}"
TEI_NS = "{http://www.tei-c.org/ns/1.0}"


def fetch_ns():
    return {"tei_ns": TEI_NS,
            "xml_ns": XML_NS}


def elem_iter(root, ns="{http://www.tei-c.org/ns/1.0}"):
    """
    Return an iterator of the elements (utterances, notes, pbs) in a protocol body

    Args:
        root (lxml.etree.Element): the protocol data as an lxml tree root
        ns (str): TEI namespace, defaults to TEI v1.0.

    """
    for body in root.findall(".//" + ns + "body"):
        for div in body.findall(ns + "div"):
            for ix, elem in enumerate(div):
                if elem.tag == ns + "u":
                    yield "u", elem
                elif elem.tag == ns + "note":
                    yield "note", elem
                elif elem.tag == ns + "pb":
                    yield "pb", elem
                #elif elem.tag == ns + "seg": # Code doesn't return segs anyway (2023-09-20), but
                #    yield "seg", elem        # commenting out in case of catastrophy -- fully delete after
                elif elem.tag == "u":
                    elem.tag = ns + "u"
                    yield "u", elem
                else:
                    LOGGER.warning(f"Encountered unknown tag: {elem.tag}")
                    yield None


def infer_metadata(filename):
    """
    Heuristically infer metadata from a protocol id or filename.

    Args:
        filename (str): the protocols filename. Agnostic wrt. dashes and underscores. Can be relative or absolute.

    Returns a dict with keys "protocol", "chamber", "year", and "number"
    """
    metadata = dict()
    filename = filename.replace("-", "_")
    metadata["protocol"] = filename.split("/")[-1].split(".")[0]
    split = filename.split("/")[-1].split("_")
    metadata["document_type"] = split[0]

    # Year
    for s in split:
        yearstr = s[:4]
        if yearstr.isdigit():
            year = int(yearstr)
            if year > 1800 and year < 2100:
                metadata["year"] = year
                metadata["sitting"] = str(year)

                # Protocol ids of format 197879 have two years, eg. 1978 and 1979
                if s[4:6].isdigit() or "urtima" in metadata["protocol"]:
                    metadata["secondary_year"] = year + 1
                    metadata["sitting"] += f"{s[4:6]}"

    # Chamber
    metadata["chamber"] = "Enkammarriksdagen"
    if "_ak_" in filename:
        metadata["chamber"] = "Andra kammaren"
    elif "_fk_" in filename:
        metadata["chamber"] = "Första kammaren"

    try:
        metadata["number"] = int(split[-1])
    except:
        pass  # print("Number parsing unsuccesful", filename)

    return metadata


def clean_html(s):
    """
    Read a HTML file and turn it into valid XML

    Args:
        s (str): original HTML as a string

    Returns:
        tree (lxml.etree.tree): XML tree
    """
    soup = BeautifulSoup(s)
    pretty_html = soup.prettify()
    return etree.fromstring(pretty_html)


def validate_xml_schema(xml_path, schema_path, schema=None):
    """
    Validate an XML file against a schema.

    Args:
        xml_path (str): path to the XML file
        schema_path (str): path to the schema file

    Returns:
        is_valid (bool): whether the XML is valid according to the schema
    """
    xml_file = etree.parse(xml_path)
    xml_file.xinclude()

    if schema is None:
        schema = xmlschema.XMLSchema(schema_path)

    try:
        schema.validate(xml_file)
        return True
    except Exception as err:
        print(err)
        return False

    return is_valid


def corpus_iterator(document_type, corpus_root=None, start=None, end=None):
    """
    Returns an iterator of document paths in a corpus.

    Args:
        document_type (str): type of document (prot, mot, etc.). Valid types:
                            - records | prot
                            - motions | mot
                            - interpellations | ipq
        corpus_root (str): path to the corpus root. If None, the function looks for the default location (see get_data_location()) based on the doctype
        start (int): start year
        end (int): end year

    Returns:
        iterator of the protocols as relative paths to current location
    """
    doctypes = {
        "interpellations": "interpellations",
        "ipq": "interpellations",
        "motions":"motions",
        "mot": "motions",
        "prot": "records",
        "records": "records"
    }
    if document_type not in doctypes:
        raise ValueError(f"{document_type} not valid")
    if corpus_root is None:
        corpus_root = get_data_location(doctypes[document_type])
    folder = Path(corpus_root)
    if folder.is_absolute():
        folder = folder.relative_to(Path(".").resolve(), walk_up=True)
    docs = folder.glob("*/*.xml")
    for doc in sorted(docs):
        if start is not None:
            metadata = infer_metadata(doc.name)
            if "year" not in metadata:
                continue
            year = metadata["year"]
            secondary_year = metadata.get("secondary_year", year)
            if start <= year and end >= secondary_year:
                yield str(doc.relative_to("."))
        else:
            yield str(doc.relative_to("."))


def protocol_iterators(corpus_root=None, document_type=None, start=None, end=None):
    """
    Deprecate - Use corpus_iterator() instead.
    Returns an iterator of protocol paths in a corpus.

    Args:
        corpus_root (str): path to the corpus root. If env variable RECORDS_PATH exists, uses that as a default
        document_type (str): type of document (prot, mot, etc.). If None, fetches all types
        start (int): start year
        end (int): end year

    Returns:
        iterator of the protocols as relative paths to current location
    """
    warnings.warn("protocol_iterators is replaced by corpus_iterator() and may be removed in future versions -- use that instead.", DeprecationWarning, stacklevel=2)
    return corpus_iterator("prot", corpus_root=corpus_root, start=start, end=end)


def parse_date(s):
    """
    Parse datetimes with special error handling

    Args:
        s (str): datetime as a string

    Returns:
        date (datetime.datetime): date as a datetime
    """
    try:
        return datetime.strptime(s, "%Y-%m-%d")

    except ValueError:
        if len(s) == 4:
            if int(s) > 1689 and int(s) < 2261:
                return datetime(int(s), 6, 15)
            else:
                return None
        else:
            return None


def get_formatted_uuid(seed=None):
    """
    Generate a UUID and format it in base58.
    The formatted UUID is prepended bu 'i-' so that it can be used as an XML ID

    Args:
        seed (str): Random seed. Optional.

    Returns:
        id (str): formatted UUID
    """
    if seed is None:
        x = uuid.uuid4()
    else:
        m = hashlib.md5()
        m.update(seed.encode('utf-8'))
        x = uuid.UUID(m.hexdigest())

    return f"i-{str(base58.b58encode(x.bytes), 'UTF8')}"


def _download_with_progressbar(url, fname, chunk_size=1024):
    resp = requests.get(url, stream=True)
    total = int(resp.headers.get('content-length', 0))
    with open(fname, 'wb') as file, tqdm(
        desc=fname,
        total=total,
        unit='iB',
        unit_scale=True,
        unit_divisor=1024,
    ) as bar:
        for data in resp.iter_content(chunk_size=chunk_size):
            size = file.write(data)
            bar.update(size)


def download_corpus(path="./", partitions=["records"]):
    """
    Downloads the full corpus.
    Does not return anything, just downloads the corpus ZIP and unzips it.

    Args:
        path (str): path for the download
    """
    p = Path(path)
    for partition in partitions:
        url = f"https://github.com/swerik-project/riksdagen-{partition}/releases/latest/download/{partition}.zip"
        zip_path = p / f"{partition}.zip"
        corpus_path = p / "data"
        if corpus_path.exists():
            LOGGER.warning(f"data already exists at the path '{corpus_path}'. It will be overwritten once the download is finished.")

        zip_path_str = str(zip_path.relative_to("."))
        extraction_path = str(p.relative_to("."))

        # Download file and display progress
        _download_with_progressbar(url, zip_path_str)
        with zipfile.ZipFile(zip_path_str, "r") as zip_ref:
            LOGGER.debug(f"Extract to {corpus_path} ...")
            zip_ref.extractall()

        zip_path.unlink()


def get_doc_dates(protocol):
    """
    Gets the content of <docDate> elements.

    Args:
        protocol: str or etree.Element

    Returns:

        match_error (bool):  True when the value of the "when" attribte doesn't match the element's text value.
        dates (list): a list of dates.
    """
    match_error = False
    dates = []
    if type(protocol) == str:
        root, ns = parse_tei(protocol)
    elif type(protocol) == etree._Element:
        root = protocol
        ns = fetch_ns()
    else:
        raise TypeError(f"You need to pass a string or etree Element, not {type(protocol)}")
    date_elems = root.findall(f".//{ns['tei_ns']}docDate")
    for de in date_elems:
        when_attrib = de.get("when")
        elem_text = de.text
        if not when_attrib == elem_text:
            match_error = True
        dates.append(when_attrib)
    return match_error, dates


def write_tei(elem, dest_path) -> None:
    """
    Write a corpus document to disk.

    Args:
        elem (etree._Element): tei root element
        dest_path (str): protocol path
    """
    elem = format_texts(elem, padding=10)
    b = etree.tostring(
        elem,
        pretty_print=True,
        encoding="utf-8",
        xml_declaration=True
    )
    with open(dest_path, "wb") as f:
        f.write(b)


def write_protocol(prot_elem, prot_path) -> None:
    """
    Write the protocol to a file.

    Args:
        prot_elem (etree._Element): protocol root element
        prot_path (str): protocol path
    """
    warnings.warn("write_protocol is replaced by write_tei() and may be removed in future versions -- use that instead.", DeprecationWarning, stacklevel=2)
    write_tei(prot_elem, prot_path)


def parse_tei(_path, get_ns=True) -> tuple:
    """
    Parse a protocol, return root element (and namespace defnitions).

    Args:
        _path (str): path to tei-xml doc
        get_ns (bool): also return namespace dict

    Returns:
        tuple/etree._Element: root and an optional namespace dict
    """
    parser = etree.XMLParser(remove_blank_text=True)
    root = etree.parse(_path, parser).getroot()
    if get_ns:
        ns = fetch_ns()
        return root, ns
    else:
        return root


def parse_protocol(protocol_path, get_ns=False) -> tuple:
    """
    Parse a protocol, return root element (and namespace defnitions).

    Args:
        protocol_path (str): protocol path
        get_ns (bool): also return namespace dict

    Returns:
        tuple/etree._Element: root and an optional namespace dict
    """
    warnings.warn("parse_protocol is replaced by parse_tei() and may be removed in future versions -- use that instead.", DeprecationWarning, stacklevel=2)
    return parse_tei(protocol_path, get_ns=get_ns)


def get_data_location(partition):
    """
    Get data location for a specific path. Tries the env variables
    RECORDS_PATH, MOTIONS_PATH and METADATA_PATH. If those do not exist
    returns the defaults data/, data/, data/
    """
    d = {}
    d["records"] = os.environ.get("RECORDS_PATH", "data")
    d["motions"] = os.environ.get("MOTIONS_PATH", "data")
    d["motions-alto"] = os.environ.get("MOTIONS_ALTO_PATH", "data")
    d["metadata"] = os.environ.get("METADATA_PATH", "data")
    d["metadata_db"] = os.environ.get("METADATA_DB", "data")               # path to csv or pkl of compiled Corpus()
    d["interpellations"] = os.environ.get("INTERPELLATIONS_PATH", "data")
    assert partition in d, f"Provide valid partition of the dataset ({list(d.keys())})"
    return d[partition]


def get_gh_link(_file,
                elem=None,
                line_number=None,
                username='swerik-project',
                repo='riksdagen-records',
                branch='dev'):
    """
    return a formatted github link to an lxml element or specified line in _file

    Args:
    - _file: the path (from root) of the file in question
    - elem: the element
    - username: the username of the repo owner
    - reponame: the repository containing the _file
    - branch: the branch you want to link to

    Returns:
        gh (str): formatted github link
    """
    if not (elem is not None or line_number is not None) and elem != line_number:
        raise ValueError("You have to pass an elem or a line number")

    if _file.startswith(repo):
        _file = _file.replace(f"{repo}/", "")
    if elem is not None:
        line_number = elem.sourceline
    gh = f"https://github.com/{username}/{repo}/blob/{branch}/{_file}/#L{line_number}"
    return gh


def remove_whitespace_from_sequence(text):
    """
    Remove repeated whitespace and replace all whitespace with spaces
    Input is string and output is string.
    """
    text_seq = text.split()
    text_seq = [s for s in text_seq if s != '']
    return ' '.join(text_seq)

  
def get_sequence_from_elem_list(elem_list):
    """
    Get sequence from first elem in list.
    Returns string. If list is empty, returns empty string. 
    """
    if len(elem_list) > 0:
        return str(elem_list[0].text)
    return ""

  
def extract_context_sequence(elem, context_type, target_length = 128, separator = '/n'):
    """
    Get sequence with context from xml element. Returns string. 
    """
    sequence_to_list_by_punctuation = lambda sequence_string: list(filter(None, re.split(r'([.!?])', sequence_string)))
    
    current_sequence = remove_whitespace_from_sequence(elem.text)
    
    previous_elem_list = elem.xpath("preceding::*[local-name() = 'note' or local-name() = 'seg'][1]")
    previous_sequence = remove_whitespace_from_sequence(get_sequence_from_elem_list(previous_elem_list))
    previous_sequence_as_list = sequence_to_list_by_punctuation(previous_sequence)
    previous_last_sentence = ''.join(previous_sequence_as_list[-2:]).lstrip('.!? ')
    
    if context_type == 'left_context':
        max_previous_length = target_length//2
    elif context_type == 'full_context':
        max_previous_length = target_length//3
        next_elem_list = elem.xpath("following::*[local-name() = 'note' or local-name() = 'seg'][1]")
        next_sequence = remove_whitespace_from_sequence(get_sequence_from_elem_list(next_elem_list))
        next_sequence_as_list = sequence_to_list_by_punctuation(next_sequence)
        next_first_sentence = ''.join(next_sequence_as_list[:2])
    
    previous_last_sentence = ' '.join(previous_last_sentence.split(' ')[-max_previous_length:]) # truncate sequence if too long
    left_context_sequence = previous_last_sentence + f' {separator} ' + current_sequence

    if context_type == 'left_context':
        return left_context_sequence
    elif context_type == 'full_context':
        return left_context_sequence + f' {separator} ' + next_first_sentence

 
def get_context_sequences_for_protocol(protocol, context_type, target_length = 128, separator = '/n'):
    """
    Gets context sequences for a protocol. Returns dictionary with ids and corresponding context sequences. 
    """
    id_list, texts_with_contexts = [], []
    
    parser = etree.XMLParser(remove_blank_text=True)
    root = etree.parse(protocol, parser).getroot()
    
    for tag, elem in elem_iter(root):
        if tag == 'note':
            elem_id = elem.get(f'{XML_NS}id')
            id_list.append(elem_id)
            context_sequence = extract_context_sequence(elem, context_type = context_type, target_length = target_length, separator = separator)
            texts_with_contexts.append(context_sequence)
        elif tag == 'u':
            for child in elem:
                child_id = child.get(f'{XML_NS}id')
                id_list.append(child_id)
                context_sequence = extract_context_sequence(child, context_type=context_type, target_length = target_length, separator = separator)
                texts_with_contexts.append(context_sequence)
    
    output_dict = {'id' : id_list,
                   'text' : texts_with_contexts}
    return output_dict


def pathize_protocol_id(protocol_id):
    """
    Turn the protocol id into a path string from riksdagen-records root. Handles 'unpadded' protocol numbers.
    """

    spl = protocol_id.split('-')
    parliament_year = spl[1]
    suffix = ""
    if len(spl) == 4:
        nr = spl[3]
        pren = '-'.join(spl[:3])
    else:
        nr = spl[5]
        pren = '-'.join(spl[:5])
        if len(spl) == 7:
            suffix = f"-{spl[-1]}"
    path_ = f"data/{parliament_year}/{pren}-{nr:0>3}{suffix}.xml"
    if os.path.exists(path_):
        # if path_ exists, return path_
        return path_
    else:
        # try to remove strings like "extra" and "höst" from path_ ...
        #    (these were removed from some of the protocol filenames after goldstandard annotation)
        #    check again if new path_ exists and return
        path_ = re.sub(f'((extra)?h[^-]+st|")', '', path_)
        if os.path.exists(path_):
            return path_
    raise FileNotFoundError(f"Can't find {path_}")



"""
Functions relating to the Swerik Catalog.
"""
import json
import pandas as pd


def _json_template(i:str, v:str, now:str) -> dict:
    """
    Returns mostly empty dict with person info.

    Args:
        i (str): person ID
        v (str): version
        now (str): string-formatted timestamp

    Returns:
        dict: person dict

    """
    return {
        "name": None,             #
        "swerik-id":i,            #
        "DOB": None,              #
        "DOD":None,               #
        "gender": None,           #
        "place-of-birth": {       #
                "place": None,    #
                "link": None      #
            },                    #
        "place-of-death": {       #
                "place": None,    #
                "link": None      #
            },                    #
        "alt-names": [],          #
        "iorter": [],             #
        "positions": [],          #
        "portraits": [],          #
        "identifiers": [],        #
        "party-affiliations":[],  #
        "data-version": v,        #
        "last-updated": now       #
    }


def _party_template(row:pd.core.series.Series) -> dict:
    """
    Returns instance of party affiliation.

    Args:
        row (pd.core.series.Series): a row

    Returns:
        dict: info about party affiliation
    """
    return {
        "party": {
            "name": row["party"],
            "party-id": row["party_id"]  # current wiki_id, will change, formatter url in template
        },
        "start": str(row["start"]),
        "end": str(row["end"]),
    }


def _position_template(row:pd.core.series.Series) -> dict:
    """
    Returns instance of a position held

    Args:
        row (pd.core.series.Series): a row

    Returns:
        dict: info about position
    """
    return {
            'start': str(row["start"]),
            'end': str(row["end"]),
            'role': row["role"],
            "source": row["source"],
            'chamber': row["chamber"],
            'government': row["government"],
            "district": row["district"]
        }


def _identifier_template(auth_code, identifier:str) -> dict:
    """
    Returns instance of an external identifier

    Args:
        auth_code: authority abbreviation
        identifier (str): an external identifier

    Returns:
        dict: info about external identifiers.
    """
    url_formatter = {
            "DiSweNaBiID": {                                                  # P3217
                    "formatter-url": "https://sok.riksarkivet.se/sbl/Presentation.aspx?id={}",
                    "name": "Dictionary of Swedish National Biography ID"
                },
            "RiPeID": {                                                       # P4342
                    "formatter-url": "https://data.riksdagen.se/personlista/?iid={}&utformat=html",
                    "name": "Riksdagen person-ID"
                },
            "SwePaPeGUID": {                                                  # P8388
                    "formatter-url": "https://www.riksdagen.se/sv/ledamoter-partier/ledamot/_{}",
                    "name": "Swedish Parliament person GUID"
                },
            "SwePoArID": {                                                    # P4819
                    "formatter-url": "https://portrattarkiv.se/details/{}",
                    "name": "Swedish Portrait Archive ID"
                },
            "UpUnAlID": {                                                     # P6821
                    "formatter-url": "https://www.alvin-portal.org/alvin/view.jsf?pid={}",
                    "name": "Uppsala University Alvin ID"
                },
            "WiDaID": {
                    "formatter-url": "https://www.wikidata.org/wiki/{}",
                    "name": "Wikidata"
                }
        }
    return {
            "authority": url_formatter[auth_code]['name'],
            "identifier": identifier,
            "link": url_formatter[auth_code]["formatter-url"].format(identifier)
        }


def _verify_J(J:dict) -> bool:
    """
    Verifies that Dict object is not totally empty before writing to a file.

    Args:
        J (dict): dictionary object

    Returns:
        bool: J is not empty
    """
    not_empty = False
    if J['name'] == None or len(J['name']) == 0:
        if J['alt-names'] == None or len(J['alt-names']) == 0:
            return not_empty
        else:
            J['name'] = J['alt-names'][0]
            not_empty = True
            return True
    for _ in ["DOB",
            "DOD",
            "gender",
            "place-of-birth",
            "place-of-death",
            "alt-names",
            "iorter",
            "positions",
            "portraits",
            "identifiers",
            "party-affiliations",
        ]:
        if J[_] is not None and len(J[_]) > 0:
            not_empty = True
            return True
    return not_empty


def _add_party(r:pd.core.series.Series, J:dict) -> bool:
    """
    Checks whether a party affiliation can be added to a person dict

    Args:
        r (pd.core.series.Series): row
        J (dict): Person dict

    Returns:
        bool: can add
    """
    if (
        pd.notnull(r['start'])
        and pd.notnull(r['end'])
        and r['party'] not in [_['party']['name'] for _ in J['party-affiliations']]
        ):
        return True
    else:
        return False


def jsonize_person_data(
        person_id:str,
        Corpus_metadata:pd.core.frame.DataFrame,
        peripheral_metadata:dict,
        v:str,
        now:str) -> dict:
    """
    Return a json object of all a n individual's info.

    Args:
        person_id (str): person ID
        Corpus_metadata (pd.core.frame.DataFrame): corpus metadata (pyriksdagen.metadata.Corpus)
        peripheral_metadata (dict): metadata dict from non-core queries
        v (str): data version
        now (str): string formatted timestamp

    Returns:
        dict: Person dict object
    """
    C = Corpus_metadata                 # for convenience
    J = _json_template(person_id, v, now)

    # process core metadata
    primary_name = C.loc[C["primary_name"] == True, 'name'].unique()
    if len(primary_name) == 1:
        J["name"] = primary_name[0]
    elif len(primary_name) == 0:
        print(f">>{person_id} -- NO PRIMARY NAME")
    else:
        print(f">>{person_id} -- MULTIPLE PRIMARY NAMES : {len(primary_name)} : {primary_name}")

    born = C.loc[pd.notnull(C["born"]), "born"].unique()
    if len(born) > 1:
        print(f">>{person_id} -- MULTIPLE BIRTHDATES : {len(born)} : {born}")
        J["DOB"] = str(born[0])
    elif len(born) == 1:
        J["DOB"] = str(born[0])

    dead = C.loc[pd.notnull(C["dead"]), "dead"].unique()
    if len(dead) > 1:
        print(f">>{person_id} -- MULTIPLE DEATHDATES : {len(dead)} : {dead}")
        J["DOD"] = str(dead[0])
    elif len(dead) == 1:
        J["DOD"] = str(dead[0])

    gender = C.loc[pd.notnull(C["gender"]), "gender"].unique()
    if len(gender) > 1:
        print(f"MULTIPLE GENDERS : {len(gender)} : {gender}")
        J["gender"] = gender[0]
    elif len(gender) == 1:
        J["gender"] = gender[0]

    alt_names = C.loc[C["primary_name"] == False, "name"].unique()
    for alt_name in alt_names:
        if alt_name != J["name"]:
            J["alt-names"].append(alt_name)

    iorter = C.loc[pd.notnull(C["location"]), "location"].unique()
    for iort in iorter:
        J["iorter"].append(iort)

    positions = C.drop_duplicates(['start', 'end', 'role', "source", 'chamber', 'government'])
    if positions is not None and not positions.empty:
        positions = positions.sort_values(by="start", ascending=False).copy()
        positions = positions.drop_duplicates().copy()
        for i, r in positions.iterrows():
            J['positions'].append(_position_template(r))

    # process periperal metadata
    PA = peripheral_metadata["party_affiliation"]
    if PA is not None and len(PA) > 0:
        PA.sort_values(by="start", ascending=False, inplace=True)
        PA = PA.drop_duplicates().copy()
        for i, r in PA.iterrows():
            if _add_party(r, J):
                J["party-affiliations"].append(_party_template(r))

    EI = peripheral_metadata["external_identifiers"]
    if EI is not None and len(EI) > 0:
        for i, r in EI.iterrows():
            J["identifiers"].append(_identifier_template(r["authority"], r['identifier']))

    PB = peripheral_metadata["place_of_birth"]
    if PB is not None and len(PB) > 0:
        J["place-of-birth"]["place"] = PB.at[0, "place"]
        J["place-of-birth"]["link"] = PB.at[0, "link"]

    PD = peripheral_metadata["place_of_death"]
    if PD is not None and not PD.empty:
        J["place-of-death"]["place"] = peripheral_metadata["place_of_death"].at[0, "place"]
        J["place-of-death"]["link"] = peripheral_metadata["place_of_death"].at[0, "link"]

    PO = peripheral_metadata["portraits"]
    if PO is not None and len(PO) > 0:
        for i, r in peripheral_metadata["portraits"].iterrows():
            J["portraits"].append(r["portrait"])

    # populate info into J
    if _verify_J(J):
        return J
    else:
        return None


def write_J(J, person_id, website_path) -> None:
    """
    Write a json file for each swerik ID with person's info.

    Args:
        J (dict): person dict
        person_id (str): person id
        website_path (str): path in website repo
    """
    with open(f"{website_path}_data/_person-catalog/{person_id}.json", "w+") as outj:
        json.dump(J, outj, ensure_ascii=False, indent=2)


def write_md(person_id:str, website_path:str) -> None:
    """
    Write markwown file for each swerik ID

    Args:
        person_id (str): person id
        website_path (str): path in website repo

    """
    markdown = f"""---
layout: catalog
title: SWERIK Person Catalog
---
{{% assign p = site.data._person-catalog.{person_id} %}}
{{% include person-catalog.html p=p %}}

"""
    with open(f"{website_path}_person-catalog/{person_id}.md", "w+") as md:
        md.write(markdown)



import numpy as np
import pandas as pd
import re
import textdistance
from unidecode import unidecode
from nltk.metrics.distance import edit_distance
from trainerlog import get_logger
LOGGER = get_logger("match_mp")

def multiple_replace(text, i_start=192, i_end=383):
	d = [chr(c) for c in range(i_start, i_end+1)]
	d = {c:unidecode(c) for c in d if c not in 'åäöÅÄÖ'}
	regex = re.compile("(%s)" % "|".join(map(re.escape, d.keys())))
	return regex.sub(lambda mo: d[mo.string[mo.start():mo.end()]], text) 

def clean_names(names):
	if type(names) == str:
		names = names.replace(',','') # Allard, Henry --> Allard Henry
		names = names.replace('.','') # A.C. Lindblad --> AC Lindblad
		names = names.replace('-',' ') # Carl-Eric Lindblad --> Carl Eric Lindblad
		names = names.lower()
	else:
		names = names.str.replace(',','', regex=False) # Allard, Henry --> Allard Henry
		names = names.str.replace('.','', regex=False) # A.C. Lindblad --> AC Lindblad
		names = names.str.replace('-',' ', regex=False) # Carl-Eric Lindblad --> Carl Eric Lindblad
		names = names.str.lower()
	return names

def name_equals(name, db):
	matches = db[db["name"] == name]
	return matches

def name_almost_equals(name, db):
	def _rough_equality(s1, s2):
		perfect_matches = 0
		s1, s2 = s1.replace(".", ""), s2.replace(".", "")
		l1 = s1.split()
		l2 = s2.split()
		if len(l1) != len(l2):
			return False
		for w1, w2 in zip(l1, l2):
			shorter_len = min(len(w1), len(w2))
			w1, w2 = w1.lower(), w2.lower()
			if w1 == w2:
				perfect_matches += 1
			if edit_distance(w1, w2) <= 1:
				if w1[0] == w2[0]:
					perfect_matches += 1
			elif w1[:shorter_len] != w2[:shorter_len]:
				return False
		return perfect_matches >= 1

	matches = db[db["name"].apply(lambda x: _rough_equality(name, x))]
	return matches

def names_in(name, db):
	names = name.split()
	matches = db[db["name"].apply(lambda x:
		len([n for n in names if n in x.split()]) == len(names))]
	return matches

def names_in_rev(name, db):
	names = name.split()
	matches = db[db["name"].apply(lambda x:
		len([n for n in x.split() if n in names]) == len(x.split()))]
	return matches

def fuzzy_name(name, db):
	d = textdistance.levenshtein
	matches = db[db["name"].apply(lambda x: d(name, x) == 1)]
	return matches

### Depreceated functions below
# NEW functions to replace both in_name, subnames_in_mpname, and mpsubnames_in_name
def subnames_in_mpname(name, db):
	if debug:
		print(name, len(name))
	indices = [i for i,row in db.iterrows() if
			   all(any(n == subname for n in row["name"].split())
			   for subname in name.split())]
	return db.loc[indices]

def mpsubnames_in_name(name, db):
	if debug:
		print(name, len(name))
	indices = [i for i,row in db.iterrows() if
			   all(any(n == subname for n in name.split())
			   for subname in row["name"].split())]
	return db.loc[indices]

def firstname_lastname(name, db):
	if debug:
		print(name, len(name))
	if len(subnames := name.split()) <= 1: return []
	indices = [i for i,row in db.iterrows() \
	if subnames[0] == row["name"].split()[0] and subnames[-1] == row["name"].split()[-1]]
	return db.loc[indices]

def firstname_lastname_reversed(name, db):
	if debug:
		print(name, len(name))
	if len(subnames := name.split()) <= 1: return []
	indices = [i for i,row in db.iterrows() \
	if subnames[0] == row["name"].split()[-1] and subnames[-1] == row["name"].split()[0]]
	return db.loc[indices]

def two_lastnames(name, db):
	if debug:
		print(name, len(name))
	if len(subnames := name.split()) <= 1: return []
	indices = [i for i,row in db.iterrows() \
	if name.split()[-1] == row["name"].split()[-1] and name.split()[-2] == row["name"].split()[1:]]
	return db.loc[indices]

def lastname(name, db):
	if debug:
		print(name, len(name))
	return db[db["name"].str.split().str[-1] == name]

def match_mp(person, db, variables, matching_funs):
	"""
	Pseudocode:
	- inputs:
		- person with cleaned name and party_abbrev
		- db with cleaned name and filtered by chamber for efficiency
		- variables list of lists with variable combinations to match by
		- matching_funs list of functions to match/filter names by
	- check if it is a talman/minister and use other matching function if true
	- filter db by gender if persons gender is available
	- use matching_funs in combination with variables until a unique match is found
	- if there at any step is a match, return the persons db id

	"""
	p = person.copy() # avoids spooky behaviour
	p["name"] = clean_names(p.get("name", ""))
	for key in ["name", "party_abbrev", "specifier"]:
		if key not in p:
			p[key] = ""
	for key in p:
		p[key] = [p[key]]
	p = pd.DataFrame.from_dict(p)
	p = p.astype(str)
	p = p.iloc[0]

	# statskalender file currently lacks gender
	if 'gender' in p:
		db = db[db["gender"] == p["gender"]]

	for fun in matching_funs:
		matched_mps = fun(p["name"], db)
		
		if len(matched_mps) == 0:
			if fun == matching_funs[-1]:
				LOGGER.debug("Intro not matched after all functions")
				return
			continue # restart if no match was found
		
		number_of_matches = len(set(matched_mps["id"]))
		if number_of_matches == 1:
			matched_value = matched_mps["id"].iloc[0]
			LOGGER.debug(f"Match found with {fun}: {matched_value}")
			return matched_value
		elif number_of_matches >= 2:
			LOGGER.debug(f"Multiple matches obtained with {fun}")

		# Iterates over combinations of variables to find a unique match
		for v in variables:
			
			matched_mps_new = matched_mps.iloc[np.where(matched_mps[v] == p[v])[0]]

			number_of_matches = len(set(matched_mps_new["id"]))
			if number_of_matches == 1:
				matched_value = matched_mps_new["id"].iloc[0]
				LOGGER.debug(f"Match found with {v}: {matched_value}")
				return matched_value
			elif number_of_matches >= 2:
				LOGGER.debug(f"Multiple matches obtained with {v}")
			



from lxml import etree
import re
from pyparlaclarin.read import element_hash
import dateparser
import pandas as pd
from .utils import elem_iter, infer_metadata, parse_date, XML_NS, TEI_NS, get_formatted_uuid
from .db import load_expressions, filter_db, load_patterns, load_metadata
from .segmentation import (
    detect_mp,
    detect_minister,
    detect_speaker,
    expression_dicts,
    detect_introduction,
    intro_to_dict,
)
from .match_mp import multiple_replace
from datetime import datetime
from trainerlog import get_logger
LOGGER = get_logger("refine")


def redetect_protocol(metadata, protocol):
    """
    For each intro in a protocol, detect which MP is mentioned and map it to metadata.

    Args:
        metadata (dict): basic metadata on the protocol
        protocol (lxml.etree): protocol as an lxml tree
    
    Returns:
        protocol as an lxml tree
    """
    tei_ns = ".//{http://www.tei-c.org/ns/1.0}"
    parser = etree.XMLParser(remove_blank_text=True)

    party_mapping, mp_db, minister_db, speaker_db = metadata
    ## DEPRECIATED ##join_intros = join_intros[join_intros['protocol'] == protocol]
    
    protocol_id = protocol.split("/")[-1]
    metadata = infer_metadata(protocol)
    root = etree.parse(protocol, parser).getroot()
    
    # Year from the folder name
    year = metadata["year"]
    # Take into account folders such as 198889
    secondary_year = metadata.get("secondary_year", year)

    dates = [
        parse_date(elem.attrib.get("when"))
        for elem in root.findall(tei_ns + "docDate")
        if parse_date(elem.attrib.get("when")).year in [year, secondary_year]
    ]
    
    # Dates from xml is wrong for digitized era        
    if len(dates) > 0:
        start_date, end_date = min(dates), max(dates)          

    else:
        start_date = datetime(year,1,1)
        end_date = datetime(secondary_year,12,31)
    
    year_mp_db = filter_db(mp_db, start_date=start_date, end_date=end_date)
    year_minister_db = filter_db(minister_db, start_date=start_date, end_date=end_date)
    year_speaker_db = filter_db(speaker_db, start_date=start_date, end_date=end_date)
    
    # Introduction patterns
    pattern_db = load_patterns()
    pattern_db = pattern_db[
        (pattern_db["start"] <= year) & (pattern_db["end"] >= year)
    ]

    root, unk = detect_mps(
        root,
        None,
        pattern_db,
        mp_db=year_mp_db,
        minister_db=year_minister_db,
        speaker_db=year_speaker_db,
        metadata=metadata,
        party_map=party_mapping,
        ## DEPRECIATED ##join_intros=join_intros,
        protocol_id=protocol_id,
        unknown_variables=["gender", "party", "other"],
    )

    b = etree.tostring(
        root, pretty_print=True, encoding="utf-8", xml_declaration=True
    )

    f = open(protocol, "wb")
    f.write(b)
    f.close()
    return unk


def detect_mps(root, names_ids, pattern_db, mp_db=None, minister_db=None, minister_db_secondary=None, speaker_db=None, metadata=None, party_map=None, protocol_id=None, unknown_variables=None):
    """
    For each intro in a protocol, detect which MP is mentioned and map it to metadata.

    Args:
        root (lxml.etree): protocol as an lxml tree
        mp_db (pd.df): MP database
        minister (pd.df): minister database 
        speaker_db (pd.df): speaker database 
        metadata (dict): basic metadata on the protocol
        party_map (dict): map from party abbreviations to party names
        join_intros (???): intros to be joined
        protocol_id (str): ID of the protocol
        unknown_variables (list): which variables to detect for unknown MPs
    
    Returns:
        protocol as an lxml tree
    """
    scanned_protocol = False
    try:
        # TODO: Fetch this from the metadata block! This solution is hacky and has broken before
        url = root.findall(".//{http://www.tei-c.org/ns/1.0}pb")[0].get('facs')
        scanned_protocol = url.startswith("https://betalab.kb.se/") or url.startswith("https://swerik-project.github.io/")
    except:
        pass

    mp_expressions = load_expressions(phase="mp")
    ## DEPRECIATED ##ids_to_join = set(join_intros['xml_id1'].tolist()+join_intros['xml_id2'].tolist())
    
    xml_ns = "{http://www.w3.org/XML/1998/namespace}"
    current_speaker = None
    prev = None
    mp_db_secondary = pd.DataFrame()

    # Extract information of unknown speakers
    unknowns = []

    #print(metadata)
    # For bicameral era, prioritize MPs from the same chamber as the protocol
    if "chamber" in metadata:
        chamber = {'Första kammaren': 1, 'Andra kammaren':2}.get(metadata['chamber'], 0)
        mp_db_secondary = mp_db[mp_db['chamber'] != chamber]
        mp_db = mp_db[mp_db['chamber'] == chamber]
        speaker_db = speaker_db[speaker_db['chamber'] == chamber]

    for tag, elem in elem_iter(root):
        parent = elem.getparent()
        if parent.attrib.get("type") != "commentSection": #ignore where people don't talk
            if tag == "u":
                # Deleting and adding attributes changes their order;
                # Mark as 'delete' instead and delete later
                elem.set("prev", "delete")
                elem.set("next", "delete")
                if current_speaker is not None:
                    elem.attrib["who"] = current_speaker
                else:
                    elem.attrib["who"] = "unknown"

                if prev is not None:
                    prev_id = prev.attrib[xml_ns + "id"]
                    elem_id = elem.attrib[xml_ns + "id"]
                    elem.set("prev", prev_id)
                    prev.set("next", elem_id)
                prev = elem

            elif tag == "note":
                if elem.attrib.get("type", None) == "speaker":
                    prev = None
                    text = elem.text
                    # Join split intros detected by BERT
                    #if elem.attrib.get(xml_ns + "id") in ids_to_join:
                    #    join_intro = join_intros.loc[
                    #                (join_intros['xml_id1'] == elem.attrib.get(xml_ns + "id")) |
                    #                (join_intros['xml_id2'] == elem.attrib.get(xml_ns + "id")), 'text']
                    #    text = join_intro.iloc[0]
                    LOGGER.debug(f"Detect intro with the text: {' '.join(text.split())}")
                    if type(text) == str:
                        d = intro_to_dict(text, mp_expressions)
                        if 'name' in d:
                            d['name'] = multiple_replace(d['name'])

                        if 'other' in d:
                            # Match minister
                            if 'statsråd' in d["other"].lower() or 'minister' in d["other"].lower():
                                current_speaker = detect_minister(text, minister_db, d)

                            elif 'talman' in d["other"].lower():
                                current_speaker = detect_speaker(text, speaker_db, metadata=metadata)

                            else:
                                current_speaker = None

                        # Match mp if not minister/talman and a name is identified
                        # if current_speaker is None and 'name' in d:
                        elif 'name' in d:

                            current_speaker = detect_mp(d, db=mp_db, party_map=party_map, match_fuzzily=scanned_protocol)

                            if current_speaker is None and len(mp_db_secondary) > 0:
                                current_speaker = detect_mp(d, db=mp_db_secondary, party_map=party_map, match_fuzzily=scanned_protocol)
                        else:
                            current_speaker = None

                        if current_speaker is None:
                            unknowns.append([protocol_id, elem.attrib.get(f'{xml_ns}id')] + [d.get(key, "") for key in unknown_variables])
        else:
            # If the whole section has no speeches, reset speaker and next/prev notation
            if tag == "u":
                elem.set("prev", "delete")
                elem.set("next", "delete")
                elem.set("who", "unknown")
            
            current_speaker = None
            prev = None

    # Do two loops to preserve attribute order
    for tag, elem in elem_iter(root):
        if tag == "u":
            if elem.attrib.get("prev") == "delete":
                del elem.attrib["prev"]
            if elem.attrib.get("next") == "delete":
                del elem.attrib["next"]

    return root, unknowns


def find_introductions(root, pattern_db, intro_ids, remove_missing=True, minister_db=None):
    """
    Find instances of curation patterns in all files in a folder.

    Args:
        root (lxml.etree): protocol as a lxml tree
        pattern_db: Patterns to be matched as a Pandas DataFrame.
        intro_ids (pd.DataFrame): List of IDs that have been detected as intros.
        remove_missing (bool): remove intros not found in 'intro_ids'
        minister_db: deprecated
    """

    # return root
    root.text = None
    current_speaker = None
    expressions, manual = expression_dicts(pattern_db)

    for ix, elem_tuple in enumerate(list(elem_iter(root))):
        tag, elem = elem_tuple
        if tag == "u":
            u = None
            u_parent = elem.getparent()
            u_parent.text = None
            for seg in list(elem):
                if type(seg.text) == str:
                    introduction = detect_introduction(
                        seg, intro_ids
                    )
                    if introduction is not None:
                        pass  # print("NEW", seg.text)
                        seg.tag = "{http://www.tei-c.org/ns/1.0}note"
                        seg.attrib["type"] = "speaker"
                        if u is not None:
                            u.addnext(seg)
                        else:
                            elem.addnext(seg)

                        u = etree.Element("{http://www.tei-c.org/ns/1.0}u")
                        # u.text = None
                        if introduction["who"] is not None:
                            u.attrib["who"] = introduction["who"]
                        else:
                            u.attrib["who"] = "unknown"

                        seg.addnext(u)
                        matched_txt = introduction["txt"]
                        ix = None
                        if matched_txt[-1] != ":" and ":" in seg:
                            ix = len(matched_txt) + seg.text.index(matched_txt)
                        if ":" in matched_txt:
                            ix = matched_txt.index(":")
                            ix = ix + seg.text.index(matched_txt)
                        elif seg.text[-1] != ":" and ":" in seg:
                            ix = seg.text.index(":")

                        if ix is not None:
                            rest = seg.text[ix + 1 :]
                            seg.text = seg.text[: ix + 1]
                            new_seg = etree.SubElement(
                                u, "{http://www.tei-c.org/ns/1.0}seg"
                            )
                            new_seg.text = rest

                    elif u is not None:
                        u.append(seg)
                        u.text = None

        elif tag == "note":
            parent = elem.getparent()
            parent.text = None
            # if not elem.attrib.get("type", None) == "speaker":
            if type(elem.text) == str:

                introduction = detect_introduction(
                    elem, intro_ids
                )

                if introduction is not None:
                    if not elem.attrib.get("type", None) == "speaker":
                        print("NEW note", elem.text)
                        elem.attrib["type"] = "speaker"

                        matched_txt = introduction["txt"]
                        ix = None
                        if matched_txt[-1] != ":" and ":" in elem.text:
                            ix = len(matched_txt) + elem.text.index(matched_txt)
                        if ":" in matched_txt:
                            ix = matched_txt.index(":")
                            ix = ix + elem.text.index(matched_txt)
                        elif elem.text[-1] != ":" and ":" in elem:
                            ix = elem.text.index(":")
                        if ix is not None:
                            rest = elem.text[ix + 1 :].strip()
                            if len(rest) > 0:
                                u = etree.Element("{http://www.tei-c.org/ns/1.0}u")
                                # u.text = None
                                if introduction["who"] is not None:
                                    u.attrib["who"] = introduction["who"]
                                else:
                                    u.attrib["who"] = "unknown"

                                elem.addnext(u)

                                rest = elem.text[ix + 1 :]
                                elem.text = elem.text[: ix + 1]

                                new_seg = etree.SubElement(
                                    u, "{http://www.tei-c.org/ns/1.0}seg"
                                )
                                new_seg.text = rest

                    else:
                        pass  # print("OLD", elem.text)
                elif elem.attrib.get("type") == "speaker" and remove_missing:
                    del elem.attrib["type"]

    return root


def detect_date(root, metadata, skip_doctors_notes=False):
    """
    Detects notes with dates in them. Updates the docDate metadata in the teiHeader accordingly.

    Args:
        root (lxml.etree): protocol as a lxml tree
        metadata (dict): basic metadata about the protocol

    Returns:
        root (lxml.etree): protocol as a lxml tree
        dates (list): list of the detected dates in an ISO format (YYYY-MM-DD)
    """

    dates = set()
    number_dates = set()
    expression = "\\w{3,5}dagen den (\\d{1,2})\\.? (\\w{3,9}) (\\d{4})"
    expression2 = "\\w{3,5}dagen den (\\d{1,2})\\.? (\\w{3,9})"
    expression3 = "(\\d{1,2})\\.? (\\w{3,9}) (\\d{4,4})"
    expression4 = "(D|d)en (\\d{1,2})\\.? (\\w{3,9})"
    protocol_year = metadata["year"]
    protocol_years = {protocol_year, metadata.get("secondary_year", protocol_year)}
    yearless = set()

    def _parse_date(s):
        return dateparser.parse(s, languages=["sv"], settings={'RETURN_AS_TIMEZONE_AWARE': False})

    def _is_sjukbetyg(elem):
        """
        check if elem with date is part of a (lakar|sjul)(be|in)tyg
        """
        if not skip_doctors_notes:
            return False
        break_while, is_sjukbetyg = False, False
        pat = re.compile(r'(läkar|sjuk)(in|be)tyg')
        while break_while == False:
            elem = elem.getprevious()
            # Run until a title or the beginning of the document is reached
            if elem is None or elem.attrib.get("type") == "title":
                break_while = True
            if elem is not None and elem.text is not None:
                m = pat.search(elem.text)
                if m is not None:
                    is_sjukbetyg = True
        return is_sjukbetyg

    for ix, elem_tuple in enumerate(list(elem_iter(root))):
        tag, elem = elem_tuple
        if tag == "note" and type(elem.text) == str and len(" ".join(elem.text.split()))  < 50:
            # Mark as to-be-removed so that attribute order is retained
            if elem.attrib.get("type") == "date":
                elem.attrib["type"] = "REMOVE"
            
            weekday_date_year = re.search(expression, elem.text)
            weekday_and_date = re.search(expression2, elem.text)
            date_and_year = re.search(expression3, elem.text)
            den_and_date = re.search(expression4, elem.text)

            # Dates with the year included, surely date
            if weekday_date_year is not None:
                if elem.attrib.get("type") != "title":
                    elem.attrib["type"] = "date"
                datestr = weekday_date_year.group(1) + " " + weekday_date_year.group(2) + " " + weekday_date_year.group(3)
                date = _parse_date(datestr)
                if date is not None and not _is_sjukbetyg(elem):
                    if date.year in protocol_years:
                        number_dates.add(date)

            # Dates with the year included, though unsure if protocol date
            elif date_and_year is not None and not _is_sjukbetyg(elem):
                datestr = date_and_year.group()
                date = _parse_date(datestr)
                if date is not None:
                    if date.year in protocol_years:
                        dates.add(date)

            # Dates without a year
            elif weekday_and_date is not None and not _is_sjukbetyg(elem):
                if elem.attrib.get("type") != "title":
                    elem.attrib["type"] = "date"
                datestr = weekday_and_date.group(1) + " " + weekday_and_date.group(2)
                yearless.add(datestr)

            # Dates without a year or weekday
            elif den_and_date is not None and not _is_sjukbetyg(elem):
                if len(" ".join(elem.text.split())) < 30:
                    if elem.attrib.get("type") != "title":
                        elem.attrib["type"] = "date"

                    datestr = den_and_date.group()
                    yearless.add(datestr)

            # Remove elements that were not dates
            if elem.attrib.get("type") == "REMOVE":
                del elem.attrib["type"]

    # Dates WITH a number are actually preferred in the hierarchy
    dates, number_dates = number_dates, dates
    if len(dates) > 0:
        protocol_year = list(dates)[0].year
    elif len(number_dates) > 0:
        dates = number_dates
        protocol_year = list(dates)[0].year

    for datestr in yearless:
        date = _parse_date(f"{datestr} {protocol_year}")
        if date is not None:
            dates.add(date)

    dates = sorted(dates)
    for text in root.findall(".//" + TEI_NS + "text"):
        for front in text.findall(".//" + TEI_NS + "front"):

            # Remove old docDates
            for docDate in front.findall(".//" + TEI_NS + "docDate"):
                docDate.getparent().remove(docDate)
            for div in front.findall(".//" + TEI_NS + "div"):
                for docDate in div.findall(".//" + TEI_NS + "docDate"):
                    docDate.getparent().remove(docDate)

            # Format dates and add default XXXX-01-01
            formatted_dates = [d.strftime("%Y-%m-%d") for d in dates]
            if len(dates) == 0:
                formatted_dates = [str(protocol_year) + "-01-01"]

            # Add new docDates
            for div in front.findall(".//" + TEI_NS + "div"):
                if div.attrib.get("type") == "preface":
                    for docDate in div.findall(".//" + TEI_NS + "docDate"):
                        docDate.getparent().remove(docDate)
                    for formatted in formatted_dates:
                        docDate = etree.SubElement(div, "docDate")
                        docDate.text = formatted
                        docDate.attrib["when"] = formatted

    return root, dates


def update_ids(root, protocol_id):
    """
    Update element IDs. TODO

    Args:
        root (lxml.etree): protocol as a lxml tree
        protocol_id (str): protocol ID. DEPRECATED

    Returns:
        root (lxml.etree): protocol as a lxml tree
        ids (set): set of all IDs in the protocol
    """
    ids = set()
    xml_id = f"{XML_NS}id"

    for tag, elem in elem_iter(root):
        if tag == "u":
            for subelem in elem:
                x = subelem.attrib.get(f'{xml_ns}id', get_formatted_uuid())
                subelem.attrib[f'{xml_ns}id'] = x
                ids.add(x)

            x = elem.attrib.get(f'{xml_ns}id', get_formatted_uuid())
            elem.attrib[f'{xml_ns}id'] = x
            ids.add(x)
                
        elif tag in ["note", "pb"]:
            x = elem.attrib.get(f'{xml_ns}id', get_formatted_uuid())
            elem.attrib[f'{xml_ns}id'] = x
            ids.add(x)

    return root, ids




#!/usr/bin/env python3
"""
Functions related to bibliographic data
"""
from tqdm import tqdm
from glob import glob

def validate_bib_data(data, bib):
    """
    Validate bibtex entries.

    Args:
        data: bibtex content
        bib: filename
    """

    OK = True
    errors = []
    if not data[0].startswith("@"):
        errors.append("First line of .bib needs to start with '@'.")
        OK = False

    try:
        bibtype, key = data[0].split("{")
        #print(filename, key[:-1], filename[:-4])
        if key[:-1] != bib.split("/")[-1][:-4]:
            errors.append("Key must be the same as the filename without '.bib'.")
            if not key.endswith(","):
                errors.append("Malformed first line. (missing comma?)")
            OK = False
    except:
        errors.append("Malformed first line.")
        OK = False

    for ix, l in enumerate(data, start=1):
        #print(ix, l)
        if ix != 1 and ix != len(data) and ix != len(data)-1:
            if not l.endswith(","):
                errors.append(f"Line {ix} (1 index) doesn't end with a comma and it should.")
                OK = False
        if ix == len(data):
            if not l == "}":
                errors.append("Last line must == '}'")
                OK = False
        if ix != 1 and ix != len(data):
            if "=" not in l:
                errors.append(f"Line {ix} (1 index) is malformed. Must contain '='.")
                OK = False
        if ix == len(data)-1:
            if l.endswith(","):
                errors.append("Second to last line shouln't end with a comma, but it does.")
                OK = False
    # More validation (planned?):
    # --- reference type (right of @) are from valid set
    # --- Fields (left of =) are from valid set

    if OK:
        return True, errors
    else:
        return False, errors


def compile_bib_list(bib_dir="./references/bibtex/", output_file="./references/compiled_references.bib"):
    """
    Compile a list of references in the references directory

    Args:
        bib_dir: path to directory containing bibtex files
        output_file: the compiled bibtex list
    """

    with open(output_file, "w+") as outf:
        for bib in tqdm(sorted(glob(f"{bib_dir}*.bib"))):
            with open(bib, 'r') as inf:
                data = [l.strip() for l in inf.readlines() if l.strip() != ""]

            valid, errors = validate_bib_data(data, bib)
            if valid:
                for idx, l in enumerate(data):
                    if idx == 0 or idx == len(data)-1:
                        outf.write(f"{l}\n")
                    else:
                        outf.write(f"    {l}\n")
                outf.write("\n\n")
            else:
                warnings.warn(f"There are some problems with bibfile: {bib}")
                [print(f"  ~~ {e}\n") for e in errors]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Class declaration and functions related to corpus metadata
"""
from functools import partial
from pyriksdagen.match_mp import multiple_replace
from pyriksdagen.utils import get_data_location
import pandas as pd
import calendar
import datetime
import os
import re




def increase_date_precision(date, start=True):
    if pd.isna(date):
        return date
    # Year
    if len(date) == 4 and start:
        return date + '-01-01'
    if len(date) == 4 and not start:
        return date + '-12-31'
    # Month
    if len(date) == 7 and start:
        return date + '-01'
    if len(date) == 7 and start:
        last_day = calendar.monthrange(int(date[0]), int(date[1]))[1]
        return date + f'-{last_day}'
    # Day
    if len(date) == 10:
        return date


def check_date_overlap(start1, end1, start2, end2):
    latest_start = max(start1, start2)
    earliest_end = min(end1, end2)
    delta = (earliest_end - latest_start).days + 1
    overlap = max(0, delta)
    return True if overlap > 0 else False


def impute_member_date(db, gov_db, from_gov='Regeringen Löfven I'):
    gov_start = gov_db.loc[gov_db['government'] == from_gov, 'start'].iloc[0]
    idx =     (db['source'] == 'member_of_parliament') &\
            (db['start'] > gov_start) &\
            (db['end'].isna())
    db.loc[idx, 'end'] = gov_db['end'].max()
    return db


def impute_member_dates(db, metadata_folder):
    def _fill_na(row, **kwargs):
        if row['start'] == 'nan' and row['end'] == 'nan':
            return row
        elif pd.isna(row['start']) and pd.isna(row['end']):
            return row
        else:
            riksmote = pd.read_csv(f"{metadata_folder}/riksdag-year.csv")
            if pd.isna(row['start']) or row['start'] == 'nan':
                try:
                    py = riksmote.loc[
                            (riksmote['start'] <= row['end']) &
                            (riksmote['end'] >= row['end'])
                        ].copy()
                    row['start'] = py['start'].unique()[0]
                except:
                    #pass
                    print("no bueno ---------------------> end:", row['end'], row['person_id'])
            elif pd.isna(row['end']) or row['end'] == 'nan':
                if int(row['start'][:4]) < 1867:
                    return row
                try:
                    py = riksmote.loc[
                            (riksmote['start'] <= row['start']) &
                            (riksmote['end'] > row['start'])
                        ].copy()
                    row['end'] = py['end'].unique()[0]
                except:
                    py = riksmote.loc[
                            riksmote['end'].str.startswith(row['start'][:4])
                        ].copy()
                    rs = sorted(py['end'].unique(), reverse=True)[0]
                    if rs < row['start']:
                        py = riksmote.loc[
                                riksmote['end'].str.startswith(
                                    str(int(row['start'][:4])+1))
                            ].copy()
                        rs = sorted(py['end'].unique(), reverse=True)[0]
                    row['end'] = rs
        return row

    def _impute_start(date, **kwargs):
        riksmote = kwargs['riksmote']
        if len(date) == 10:
            return date
        elif len(date) == 7:
            s = sorted(list(riksmote.loc[riksmote['start'].str.startswith(date, na=False), 'start']))
            if len(s) > 0:
                return s[0]
            else:
                return date + "-01"
        else:
            s = sorted(list(riksmote.loc[riksmote['start'].str.startswith(date, na=False), 'start']))
            if len(s) > 0:
                return s[0]
            else:
                print(f"Problem with start date: {date} not in riksmote")
                return date + '-01-01'

    def _impute_end(date, **kwargs):
        riksmote = kwargs['riksmote']
        if len(date) == 10:
            return date
        elif len(date) == 7:
            s = sorted(list(riksmote.loc[riksmote['end'].str.startswith(date, na=False), 'end']), reverse=True)
            if len(s) > 0:
                return s[0]
            else:
                date_year, date_month = date.split("-")
                last_day_of_the_month = calendar.monthrange(int(date_year), int(date_month))[1]
                return date + f'-{last_day_of_the_month}'
        else:
            s = sorted(list(riksmote.loc[riksmote['end'].str.startswith(date, na=False), 'end']), reverse=True)
            if len(s) > 0:
                return s[0]
            else:
                print(f"Problem with end date: {date} not in riksmote")
                return date + '-12-31'

    riksmote = pd.read_csv(f"{metadata_folder}/riksdag-year.csv")
    riksmote[['start', 'end']] = riksmote[['start', 'end']].astype(str)

    idx = (db['source'] == 'member_of_parliament') &\
            (((pd.isna(db['start'])) | (db['start'] == 'nan')) |\
            ((pd.isna(db['end'])) | (db['end'] == 'nan')))
    db.loc[idx] = db.loc[idx].apply(lambda x: _fill_na(x, riksmote=riksmote), axis = 1)

    idx = (db['source'] == 'member_of_parliament') &\
            (pd.notnull(db['start'])) & (db['start'] != 'nan')
    db.loc[idx, 'start'] = db.loc[idx, 'start'].apply(_impute_start, riksmote=riksmote)

    idx = (db['source'] == 'member_of_parliament') &\
            (pd.notnull(db['start'])) &\
            (pd.notnull(db['end']))  & (db['end'] != 'nan')
    db.loc[idx, 'end'] = db.loc[idx, 'end'].apply(_impute_end, riksmote=riksmote)
    return db


def impute_minister_date(db, gov_db):
    def _impute_minister_date(minister, gov_db):
        if pd.isna(minister['start']):
            minister['start'] = gov_db.loc[gov_db['government'] == minister['government'], 'start'].iloc[0]
        if pd.isna(minister['end']):
            minister['end'] = gov_db.loc[gov_db['government'] == minister['government'], 'end'].iloc[0]
        return minister

    # Impute missing minister dates using government dates
    if 'source' in db.columns:
        db.loc[db['source'] == 'minister'] =\
        db.loc[db['source'] == 'minister'].apply(partial(_impute_minister_date, gov_db=gov_db), axis=1)
    else:
        db = db.apply(partial(_impute_minister_date, gov_db=gov_db), axis=1)
    return db


def impute_speaker_date(db):
    if "source" in db.columns:
        idx =     (db['source'] == 'speaker') &\
                (db['end'].isna()) &\
                (db['role'].str.contains('kammare') == False)
        db.loc[idx, 'end'] = db.loc[idx, 'start'] + datetime.timedelta(days = 365*4)
    else:
        idx =     (db['end'].isna()) &\
                (db['role'].str.contains('kammare') == False)
        db.loc[idx, 'end'] = db.loc[idx, 'start'] + datetime.timedelta(days = 365*4)
    return db


def impute_date(db, metadata_folder):
    db[["start", "end"]] = db[["start", "end"]].astype(str)
    if 'source' in db.columns:
        sources = set(db['source'])
        if 'member_of_parliament' in sources:
            #db = impute_member_date(db, gov_db)
            db = impute_member_dates(db, metadata_folder)

        db['start'] = db['start'].apply(increase_date_precision, start=True)
        db['end'] = db['end'].apply(increase_date_precision, start=False)
        db[["start", "end"]] = db[["start", "end"]].apply(pd.to_datetime, format='%Y-%m-%d')

        gov_db = pd.read_csv(f'{metadata_folder}/government.csv')
        gov_db[["start", "end"]] = gov_db[["start", "end"]].apply(pd.to_datetime, format='%Y-%m-%d')
        idx = gov_db['start'].idxmax()
        gov_db.loc[idx, 'end'] = gov_db.loc[idx, 'start'] + datetime.timedelta(days = 365*4)

        if 'member_of_parliament' in sources:
            db = impute_member_date(db, gov_db)
        if 'minister' in sources:
            db = impute_minister_date(db, gov_db)
        if 'speaker' in sources:
            db = impute_speaker_date(db)

    else:
        db['start'] = db['start'].apply(increase_date_precision, start=True)
        db['end'] = db['end'].apply(increase_date_precision, start=False)
        db[["start", "end"]] = db[["start", "end"]].apply(pd.to_datetime, format='%Y-%m-%d')
    return db


def impute_party(db, party):
    if 'party' not in db.columns:
        db['party'] = pd.Series(dtype=str)
    data = []
    for i, row in db[db['party'].isnull()].iterrows():
        parties = party[party['person_id'] == row['person_id']]
        if len(set(parties['party'])) == 1:
            db.loc[i,'party'] = parties['party'].iloc[0]
        if len(set(parties['party'])) >= 2:
            for j, sow in parties.iterrows():
                try:
                    res = check_date_overlap(row['start'], sow['start'], row['end'], sow['end'])
                except:
                    print("Impute dates on Corpus using impute_date() before imputing parties!\n")
                    raise
                if res:
                    m = row.copy()
                    m['party'] = sow['party']
                    data.append(m)
    db = pd.concat([db, pd.DataFrame(data)]).reset_index(drop=True)
    return db


def abbreviate_party(db, party):
    party = {row['party']:row['abbreviation'] for _, row in party.iterrows()}
    db["party_abbrev"] = db["party"].fillna('').map(party)
    return db


def clean_name(db):
    idx = db['name'].notna()
    db.loc[idx, 'name'] = db.loc[idx, 'name'].str.lower()
    db.loc[idx, 'name'] = db.loc[idx, 'name'].astype(str).apply(multiple_replace)
    db.loc[idx, 'name'] = db.loc[idx, 'name'].str.replace('-', ' ', regex=False)
    db.loc[idx, 'name'] = db.loc[idx, 'name'].str.replace(r'[^a-zåäö\s\-]', '', regex=True)
    return db


def infer_chamber(db):
    def _infer_chamber(role):
        d = {'första': 1, 'andra': 2}
        match = re.search(r'([a-zåäö]+)\s*(?:kammar)', role)
        return d[match.group(1)] if match else 0
    db['chamber'] = db['role'].apply(_infer_chamber).astype(dtype=pd.Int8Dtype())
    return db


def format_member_role(db):
    db['role'] = db['role'].str.extract(r'(ledamot)')
    return db


def format_minister_role(db):
    db["role"] = db["role"].str.replace('Sveriges ', '').str.lower()
    return db


def format_speaker_role(db):
    def _format_speaker_role(role):
        match = re.search(r'(andre |förste |tredje )?(vice )?talman', role)
        return match.group(0)
    db['role'] = db['role'].apply(_format_speaker_role)
    return db


class Corpus(pd.DataFrame):
    """
    Store corpus metadata as a single pandas DataFrame where
    the column 'source' indicates the type of the row
    """
    def __init__(self, *args, **kwargs):
        super(Corpus, self).__init__(*args, **kwargs)

    @property
    def _constructor(self):
        return Corpus

    def _load_metadata(self, file, metadata_folder="corpus/metadata", source=False):
        df = pd.read_csv(f"{metadata_folder}/{file}.csv")

        # Adjust to new structure where party information
        # is not included in member_of_parliament.csv
        if file == "member_of_parliament":
            print(df)
            columns = list(df.columns) + ["party"]
            party_df = pd.read_csv(f"{metadata_folder}/party_affiliation.csv")
            party_df = party_df[party_df["start"].notnull()]
            party_df = party_df[party_df["end"].notnull()]
            df = df.merge(party_df, on=["person_id", "start", "end"], how="left")
            df = df[columns]
            print(df)
            print(df[df["party"].notnull()])
        if source:
            df['source'] = file
        return df

    def add_mps(self, metadata_folder="corpus/metadata"):
        df = self._load_metadata('member_of_parliament', metadata_folder=metadata_folder, source=True)
        df = infer_chamber(df)
        df = format_member_role(df)
        return Corpus(pd.concat([self, df]))

    def add_ministers(self, metadata_folder="corpus/metadata"):
        df = self._load_metadata('minister', metadata_folder=metadata_folder, source=True)
        df = format_minister_role(df)
        return Corpus(pd.concat([self, df]))

    def add_speakers(self, metadata_folder="corpus/metadata"):
        df = self._load_metadata('speaker', metadata_folder=metadata_folder, source=True)
        df = infer_chamber(df)
        df = format_speaker_role(df)
        return Corpus(pd.concat([self, df]))

    def add_persons(self, metadata_folder="corpus/metadata"):
        df = self._load_metadata('person', metadata_folder=metadata_folder)
        return self.merge(df, on='person_id', how='left')

    def add_location_specifiers(self, metadata_folder="corpus/metadata"):
        df = self._load_metadata('location_specifier', metadata_folder=metadata_folder)
        return self.merge(df, on='person_id', how='left')

    def add_names(self, metadata_folder="corpus/metadata"):
        df = self._load_metadata('name', metadata_folder=metadata_folder)
        return self.merge(df, on='person_id', how='left')

    def impute_dates(self, metadata_folder="corpus/metadata"):
        return impute_date(self, metadata_folder)

    def impute_parties(self, metadata_folder="corpus/metadata"):
        df = self._load_metadata('party_affiliation', metadata_folder=metadata_folder)
        df = impute_date(df, metadata_folder)
        return impute_party(self, df)

    def abbreviate_parties(self, metadata_folder="corpus/metadata"):
        df = self._load_metadata('party_abbreviation', metadata_folder=metadata_folder)
        return abbreviate_party(self, df)

    def add_twitter(self, metadata_folder="corpus/metadata"):
        df = self._load_metadata('twitter', metadata_folder=metadata_folder)
        return self.merge(df, on='person_id', how='left')

    def clean_names(self):
        return clean_name(self)


def load_Corpus_metadata(metadata_folder=None, read_db=False, read_db_from=None):
    """
    Populates Corpus object
    """
    if read_db or read_db_from is not None:
        if read_db_from is None:
            read_db_from = get_data_location("metadata_db")
            if not os.path.exists(read_db_from):
                raise FileNotFoundError(f"File not found at {read_db_from}. Try compiling the database or set the METADATA_DB variable in your environment.")

        print("Reading metadata db from a file.")
        try:
            corpus = pd.read_csv(read_db_from)
        except:
            corpus = pd.read_pickle(read_db_from)
        assert type(corpus) == Corpus, f"{read_db_from} is not a CSV or pickle file."
    else:
        print("Compiling metadata db from source.")
        if metadata_folder is None:
            metadata_folder = get_data_location("metadata")

        corpus = Corpus()

        corpus = corpus.add_mps(metadata_folder=metadata_folder)
        corpus = corpus.add_ministers(metadata_folder=metadata_folder)
        corpus = corpus.add_speakers(metadata_folder=metadata_folder)

        corpus = corpus.add_persons(metadata_folder=metadata_folder)
        corpus = corpus.add_location_specifiers(metadata_folder=metadata_folder)
        corpus = corpus.add_names(metadata_folder=metadata_folder)

        corpus = corpus.impute_dates(metadata_folder=metadata_folder)
        corpus = corpus.impute_parties(metadata_folder=metadata_folder)
        corpus = corpus.abbreviate_parties(metadata_folder=metadata_folder)
        corpus = corpus.add_twitter(metadata_folder=metadata_folder)
        corpus = corpus.clean_names()

        # Clean up speaker role formatting
        corpus["role"] = corpus["role"].replace({
            'Sveriges riksdags talman':'speaker',
            'andra kammarens andre vice talman':'ak_2_vice_speaker',
            'andra kammarens förste vice talman':'ak_1_vice_speaker',
            'andra kammarens talman':'ak_speaker',
            'andra kammarens vice talman':'ak_1_vice_speaker',
            'andre vice talman i första kammaren':'fk_2_vice_speaker',
            'första kammarens talman':'fk_speaker',
            'första kammarens vice talman':'fk_1_vice_speaker',
            'förste vice talman i första kammaren':'fk_1_vice_speaker'
            })

        # Temporary ids
        corpus['person_id'] = corpus['person_id']

        # Drop individuals with missing names
        corpus = corpus[corpus['name'].notna()]

        # Remove redundancy and split file
        corpus = corpus.drop_duplicates()
        #print( corpus.loc[(pd.isna(corpus['start'])) | (pd.isna(corpus['end']))] )
        corpus = corpus.dropna(subset=['name', 'start', 'end'])
        corpus = corpus.sort_values(['person_id', 'start', 'end', 'name'])

    return corpus


def fetch_person_name(person_id, corpus, primary=True):
    """
    Get a person's primary name or list of names.

    Args:
        person_id (str): a swerik-style person ID
        corpus (df): compiled metadata DB
        primary (bool): return only person's primary name

    Returns:
        name or names (str or list of str)
    """
    def _names(df):
        return df["name"].unique()

    if person_id is None or person_id == "unknown":
        return None
    df = corpus.loc[corpus["person_id"] == person_id].copy()
    if primary == True:
        df = df.loc[corpus["primary_name"] == True].copy()
        names = _names(df)
        if len(names) == 1:
            return names[0]
        else:
            raise ValueError(f"There should be exactly one primary name for {person_id}, but: {names}")
    else:
        return _names(df)


def fetch_person_gender(person_id, corpus):
    """
    Get Person's gender.

    Args:
        person_id (str): a swerik-style person ID
        corpus (df): compiled metadata DB

    Returns:
        gender (str): man or woman
    """
    if person_id is None or person_id == "unknown":
        return None
    df = corpus.loc[(corpus["person_id"] == person_id) & (pd.notnull(corpus["gender"]))].copy()
    genders = df["gender"].unique()
    if len(genders) == 1:
        return genders[0]
    elif len(genders) == 0:
        return None
    else:
        raise ValueError(f"There probably shouldn't be multiple genders for {person_id}, but {genders}")


def fetch_person_party(person_id, corpus, date=None):
    """
    Get a person's party affiliation (during a particular period.

    Args:
        person_id (str): a swerik-style person ID
        corpus (df): compiled metadata DB
        date (str): (yyyy-mm-dd) formatted date string.
            If the date arg is not None, the fn will try to match
            party affiliations that overlap with that date. If none
            are found, it defaults to all party affiliations

    Returns:
        party/ies: str or list of str
    """
    def _party(df):
        parties = df["party"].unique()
        if len(parties) == 1:
            return "string", parties[0]
        elif len(parties) == 0:
            return None
        else:
            return parties

    if person_id is None or person_id == "unknown":
        return None

    df = corpus.loc[(corpus["person_id"] == person_id) & (pd.notnull(corpus["party"]))].copy()
    if date is not None:
        df_date = df.loc[(df["start"] <= date) & (df["end"] >= date)]
        parties = _party(df_date)
        if parties is not None:
            return parties
    return _party(df)





"""
Parla Clarin generation
"""
import pandas as pd
import progressbar, copy
from lxml import etree
from pyparlaclarin.create import pc_header, create_parlaclarin
from pyparlaclarin.refine import format_texts

from .utils import infer_metadata, get_formatted_uuid, get_data_location, XML_NS, TEI_NS
from .db import year_iterator

def dict_to_tei(data):
    """
    Convert a metadata dict into a TEI XML tree

    Args:
        data (dict): dictionary containing protocol level metadata

    Returns:
        tei (lxml.etree.Element): the protocol as a TEI-formatted lxml tree root
    """
    metadata = copy.deepcopy(data)

    nsmap = {None: TEI_NS}
    nsmap = {key: value.replace("{", "").replace("}", "") for key,value in nsmap.items()}
    tei = etree.Element("TEI", nsmap=nsmap)
    protocol_id = metadata["protocol_id"]
    metadata["document_title"] = (
        f"Swedish parliamentary record {metadata['sitting']}, number {metadata['number']}"
    )
    documentHeader = pc_header(metadata)
    tei.append(documentHeader)

    text = etree.SubElement(tei, "text")
    front = etree.SubElement(text, "front")
    preface = etree.SubElement(front, "div", type="preface")
    etree.SubElement(preface, "head").text = protocol_id.split(".")[0]
    if "date" not in metadata:
        year = metadata.get("year", 2020)
        metadata["date"] = str(year) + "-01-01"

    etree.SubElement(preface, "docDate", when=metadata["date"]).text = metadata.get(
        "date", "2020-01-01"
    )

    body = etree.SubElement(text, "body")
    body_div = etree.SubElement(body, "div")

    protocol_id = protocol_id.replace("_", "-")
    element_seed = f"{protocol_id}\nNA\n"
    print(element_seed)
    for paragraph in data["paragraphs"]:
        if type(paragraph) == int:
            element_seed = f"{protocol_id}\n{paragraph}\n"
            pb = etree.SubElement(body_div, "pb")
            sitting, number = metadata["sitting"], metadata["number"]
            paragraph = f"{paragraph:03d}"
            link = f"https://betalab.kb.se/prot-{sitting}--{number}/prot_{sitting}__{number}-{paragraph}.jp2/_view"
            pb.attrib["facs"] = link
        else:
            note = etree.SubElement(body_div, "note")
            note.text = paragraph
            element_seed += paragraph
            note.attrib[f"{XML_NS}id"] = get_formatted_uuid(element_seed)

    return tei

def add_to_corpus_meta_file_list(metadata):
    """
    add entry to corpus-level metadata xml files
    """
    xi_ns = "{http://www.w3.org/2001/XInclude}"
    chambers = {
        "Enkammarriksdagen": "ek",
        "Andra kammaren": "ak",
         "Första kammaren": "fk"
    }
    parser = etree.XMLParser(remove_blank_text=True)
    records_location = get_data_location("records")
    meta_path = f"{records_location}/prot-{chambers[metadata['chamber']]}.xml"
    file_path = f"./{metadata['sitting']}/{zero_pad_prot_nr(metadata['protocol'])}.xml"
    root = etree.parse(meta_path, parser).getroot()
    include = root.find(f".//{xi_ns}include[@href=\"{file_path}\"]")
    if include:
        print(f"File already listed in corpus metadata -- {metadata['protocol']}. Are you re-curating")
    else:
        include = etree.SubElement(root, f"{xi_ns}include")
        include.attrib["href"] = file_path
    with open(meta_path, "wb") as f:
        b = etree.tostring(
            root, pretty_print=True, encoding="utf-8", xml_declaration=True
        )
        f.write(b)



def dict_to_parlaclarin(data):
    """
    Create per-protocol parlaclarin files of all files provided in file_db.
    Does not return anything, instead writes the data on disk.

    Args:
        data (dict): metadata and data

    Returns:
        None
    """
    session = data["session"]
    default_metadata = dict(
        document_title=f"Riksdagens protocols {session}",
        authority="National Library of Sweden and the WESTAC project",
        correction="Some programmatic corrections have been made to counter errors stemming from the OCR process.",
    )
    for key in default_metadata:
        if key not in data:
            data[key] = default_metadata[key]

    protocol_id = data["protocol_id"].replace("_", "-")
    number = str(data["number"])
    protocol_id = protocol_id[:-len(number)]

    number = f"{data['number']:03d}"
    protocol_id += number
    print("Formatted protocol ID", protocol_id)
    yearstr = protocol_id[5:]
    yearstr = yearstr.split("-")[0]
    records_location = get_data_location("records")
    parlaclarin_path = f"{records_location}/{yearstr}/{protocol_id}.xml"

    tei = dict_to_tei(data)
    with open(parlaclarin_path, "wb") as f:
        b = etree.tostring(
            tei, pretty_print=True, encoding="utf-8", xml_declaration=True
        )
        f.write(b)
    parser = etree.XMLParser(remove_blank_text=True)
    tei = etree.parse(parlaclarin_path, parser).getroot()
    tei = format_texts(tei, padding=10)
    with open(parlaclarin_path, "wb") as f:
        b = etree.tostring(
            tei, pretty_print=True, encoding="utf-8", xml_declaration=True
        )
        f.write(b)


def zero_pad_prot_nr(protocol_id):
    constituents = protocol_id.split('-')
    if len(constituents) > 1:
        id_nr = '{:0>3}'.format(constituents[-1])
        constituents = constituents[:-1]
        constituents.append(f"{id_nr}")
        protocol_id = '-'.join(constituents)
    else:
        constituents = protocol_id.split('_')
        id_nr = '{:0>3}'.format(constituents[-1])
        constituents = constituents[:-1]
        constituents.append(f"{id_nr}")
        protocol_id = '-'.join(constituents)
    return protocol_id




from transformers import BertTokenizerFast, AutoTokenizer
import torch

class IntroDataset(torch.utils.data.Dataset):
    ''' Dataset for predicting introductions '''
    def __init__(self, df):
        self.df = df
        self.tokenizer = BertTokenizerFast.from_pretrained('KB/bert-base-swedish-cased')

    def __len__(self):
        return len(self.df)

    def __getitem__(self, index):

        df_row = self.df.iloc[index]
        token_info = self.tokenizer.encode_plus(
                                            df_row['text'],                       # Sentence to encode.
                                            add_special_tokens = True,      # Add '[CLS]' and '[SEP]'
                                            max_length = 50,                # Pad & truncate all sentences.
                                            padding='max_length',
                                            truncation=True,
                                            return_attention_mask = True,   # Construct attn. masks.
                                            return_tensors = 'pt',          # Return pytorch tensors.
                                            )

        return token_info, df_row['id'], df_row['file_path']


class MergeDataset(torch.utils.data.Dataset):
    ''' Dataset for predicting if introduction should be merged with next coming textblock '''
    def __init__(self, df):
        self.df = df
        self.tokenizer = AutoTokenizer.from_pretrained('jesperjmb/MergeIntrosNSP')

    def __len__(self):
        return len(self.df)

    def __getitem__(self, index):

        df_row = self.df.iloc[index]
        token_info = self.tokenizer(df_row['text1'],
                                    df_row['text2'],
                                    max_length = 50,
                                    padding='max_length',
                                    truncation='longest_first',
                                    return_attention_mask = True,
                                    return_tensors="pt"
                                    )

        return token_info, df_row['xml_id1'], df_row['xml_id2'], df_row['text1'], df_row['text2'], df_row['protocol']


  


import pandas as pd
import os, json, re, hashlib
from .utils import get_data_location
from .metadata import load_Corpus_metadata
from importlib_resources import files

def year_iterator(file_db):
    """
    Iterate over triplets of (corpus_year, package_ids, year_db) for provided file database.
    """
    file_db_years = sorted(list(set(file_db["year"])))
    print("Years to be iterated", file_db_years)
    for corpus_year in file_db_years:
        year_db = file_db[file_db["year"] == corpus_year]
        package_ids = year_db["protocol_id"]
        package_ids = list(package_ids)
        package_ids = sorted(package_ids)

        yield corpus_year, package_ids, year_db


def load_patterns(year=None, phase="segmentation"):
    """
    Load regex patterns from disk
    """
    file = files(f'pyriksdagen.data.{phase}').joinpath('patterns.json')
    with file.open() as f:
        patterns = pd.read_json(f, orient="records", lines=True)
    if year is not None:
        patterns = patterns[patterns["start"] >= year]
        patterns = patterns[patterns["end"] <= year]

    patterns["protocol_id"] = None

    manual_path = "input/" + phase + "/manual.csv"
    if os.path.exists(manual_path):
        manual = pd.read_csv(manual_path)
        manual["type"] = "manual"
        return pd.concat([manual, patterns])
    else:
        return patterns


def filter_db(db, start_date=None, end_date=None, year=None, protocol_id=None):
    """
    Filter dataframe either based on year or protocol id
    """
    assert (
        year is not None or protocol_id is not None or (start_date is not None and end_date is not None)
    ), "Provide either year or protocol id"
    if start_date is not None and end_date is not None:
        filtered_db = db[(db["start"].dt.date <= end_date.date()) & (db["end"].dt.date >= start_date.date())]
        return filtered_db
    elif year is not None:
        if "start" in db.columns:
            filtered_db = db[db["start"].dt.year <= year]
            filtered_db = filtered_db[filtered_db["end"].dt.year >= year]
            return filtered_db
        elif "year" in db.columns:
            filtered_db = db[db["year"] == year]
            return filtered_db
        else:
            return None

    else:
        return db[db["protocol_id"] == protocol_id]

def load_ministers(path='corpus/wiki-data/minister.json'):
    '''Unpacks very nested minister.json file to a df.'''
    with open(path, 'r') as f:
        minister = json.load(f)

    data = []
    for gov in minister:
        g = gov["government"]
        for member in gov["cabinet"]:
            Q = member["wiki_id"]
            n = member["name"]
            for pos in member["positions"]:
                r = pos["role"]
                s = pos["start"]
                e = pos["end"]
                data.append([g, Q, n, r, s, e])
    minister = pd.DataFrame(data, columns=["government", "wiki_id", "name", "role", "start", "end"])
    return minister

def load_metadata(metadata_location=None, processed_metadata_folder=None):
    if metadata_location is None:
        metadata_location = get_data_location("metadata")
    party_mapping = pd.read_csv(f'{metadata_location}/party_abbreviation.csv')
    mb_db, minister_db, speaker_db = None, None, None
    try:
        mp_db = pd.read_csv(f'{processed_metadata_folder}/member_of_parliament.csv')
        minister_db = pd.read_csv(f'{processed_metadata_folder}/minister.csv')
        speaker_db = pd.read_csv(f'{processed_metadata_folder}/speaker.csv')
    except Exception:
        df = load_Corpus_metadata(metadata_location)        
        mp_db  = df[df['source'] == 'member_of_parliament']
        minister_db  = df[df['source'] == 'minister']
        speaker_db  = df[df['source'] == 'speaker']

    ### Temporary colname changes
    mp_db["specifier"] = mp_db["location"]
    mp_db = mp_db.rename(columns={'person_id':'id'})
    minister_db = minister_db.rename(columns={'person_id':'id'})
    speaker_db = speaker_db.rename(columns={'person_id':'id'})

    # Datetime format
    mp_db[["start", "end"]] = mp_db[["start", "end"]].apply(pd.to_datetime, errors="coerce")
    minister_db[["start", "end"]] = minister_db[["start", "end"]].apply(pd.to_datetime, errors="coerce")
    speaker_db[["start", "end"]] = speaker_db[["start", "end"]].apply(pd.to_datetime, errors="coerce")

    return party_mapping, mp_db, minister_db, speaker_db

def load_expressions(phase="segmentation", year=None):
    if phase == "segmentation":
        patterns = load_patterns(year=year)
        expressions = dict()
        for _, row in patterns.iterrows():
            pattern = row["pattern"]
            exp = re.compile(pattern)
            # Calculate digest for distringuishing patterns without ugly characters
            pattern_digest = hashlib.md5(pattern.encode("utf-8")).hexdigest()[:16]
            expressions[pattern_digest] = exp
        return expressions
    elif phase == "mp":
        file = files(f'pyriksdagen.data.segmentation').joinpath('detection.csv')
        with file.open() as f:
            patterns = pd.read_csv(f, sep=";")
        expressions = []
        for _, row in patterns.iterrows():
            exp, t = row[["pattern", "type"]]
            expressions.append((re.compile(exp), t))
        return expressions
    elif phase == "join_intros":
        file = files(f'pyriksdagen.data.segmentation').joinpath('join_intro_pattern.csv')
        with file.open() as f:
            #"input/segmentation/join_intro_pattern.csv"
            patterns = pd.read_csv(f, sep=";")
        expressions = []
        for _, row in patterns.iterrows():
            exp, t = row[["pattern", "type"]]
            expressions.append((re.compile(exp), t))
        return expressions

def _keep_most_significant(df, cols, id="wiki_id"):
    for col in cols:
        primary = df[df[col] != df[col].str[:4]]
        primary = primary[primary[col].notnull()]

        #primary = primary.drop_duplicates([id, col])
        secondary = df[df[col] == df[col].str[:4]]
        secondary = secondary[secondary[col].notnull()]

        secondary = secondary.drop_duplicates([id, col])

        col_df = pd.concat([primary, secondary])
        col_df = col_df.drop_duplicates(id)
        col_df = col_df[[id, col]]

        df = df[[c for c in df.columns if c != col]]
        df = df.drop_duplicates()
        df = pd.merge(df, col_df, how="left", on=id)

    col = cols[0]
    primary = df[df[col] != df[col].str[:4]]
    secondary = df[df[col] == df[col].str[:4]]

    df = pd.concat([primary, secondary])
    df = df.drop_duplicates(id)
    return df

def clean_person_duplicates(df):
    dupl = df[df.duplicated("person_id", keep=False)].copy()
    df = df[~df.duplicated("person_id", keep=False)]
    dupl = _keep_most_significant(dupl, ["born", "dead"], id="person_id")
    cols = list(df.columns)
    df = pd.concat([dupl, df])
    df = df[cols]
    df = df.drop_duplicates(list(df.columns))
    df = df.sort_values(list(df.columns))
    return df



#!/usr/bin/env python3
"""
Handle issues relating to dates in the corpus and MP/Minister database.
"""
from tqdm import tqdm
import pandas as pd
from pyriksdagen.metadata import (
    impute_date,            # db
    impute_minister_date,   # db gov_db
    impute_speaker_date,    # db
)
import sys



def _get_parliament_years(start, end, start_year, end_year, start_p, end_p, year_list, riksmote, debug):
    """
    get a list of parliament years in 4 or 6 digit format
    """
    parliament_years = []
    # same start/end year or end year is none
    if start_year == end_year or end_year == None:
        if str(start_year) == "1999":
            if start <= "1999-09-15":
                parliament_years.append(19981999)
            else:
                parliament_years.append(19992000)
            if end:
                if end_year > "1999":
                    parliament_years.append(19992000)
                else:
                    if end_p == "day":
                        if end >= "1999-09-15":
                            parliament_years.append(19992000)
            parliament_years.append(19992000)
        elif str(start_year) == "1975":
            if end_year:
                if end_year > "1975":
                    parliament_years.append(1975)
                    parliament_years.append(197576)
                else:
                    if end_p == "day":
                        if end > "1975-10-15":
                            parliament_years.append(1975)
                            parliament_years.append(197576)
                        else:
                            parliament_years.append(1975)
            else:
                if start_p == "day":
                    if start >= "1975-10-15":
                        parliament_years.append(197576)
                    else:
                        parliament_years.append(1975)
                else:
                    parliament_years.append(1975)

        # if an year is an acceptable parliament year
        else:
            if start_year in year_list:
                parliament_years.append(int(start_year))


            # else
            else:
                try:
                    df = riksmote.loc[riksmote["parliament_year"] == int(start_year + f"{int(str(start_year)[2:])+1:02}")]
                    #print(df)
                    assert df.empty == False
                except:
                    print("start date except")
                    print(start_year + f"{int(str(start_year)[2:])+1:02}" + " not in parliament_years.")
                else:
                    if start_p == "day":
                        df_starts =  df["start"].unique()
                        if start >= df_starts[0]:
                            parliament_years.append(int(start_year + f"{int(str(start_year)[2:])+1:02}"))
                            return parliament_years
                    else:
                        parliament_years.append(int(start_year + f"{int(str(start_year)[2:])+1:02}"))
                        return parliament_years
                try:
                    x = f"{int(str(start_year)[2:])-1:02}"
                    if x == "-1":
                        xx = 19992000
                    else:
                        xx = int(str(start_year)[:2] + x + f"{int(str(start_year)[2:]):02}")

                    df = riksmote.loc[riksmote["parliament_year"] == xx]
                    assert df.empty == False
                except:
                    print("end date except")
                    print(int(str(start_year)[:2] + x + f"{int(str(start_year)[2:]):02}"))
                    print(start, end, start_year, end_year)
                else:
                    if start_p == "day":
                        df_starts =  df["start"].unique()
                        if start >= df_starts[0]:
                            parliament_years.append(xx)
                    else:
                        parliament_years.append(xx)
    else:
        # if start year is one year before end year
        if str(int(start_year)+1) == str(end_year):
            if str(start_year) == "1999":
                parliament_years.append(19992000)
            elif str(start_year) == "1975":
                if end_year:
                    if str(end_year) > "1975":
                        if debug:
                            print("end is > 1975")
                        parliament_years.append(1975)
                        parliament_years.append(197576)
                    elif end_p == "day":
                        if end > "1975-10-15":
                            parliament_years.append(1975)
                            parliament_years.append(197576)
                        else:
                            parliament_years.append(1975)
                else:
                    if start_p == "day":
                        if start >= "1975-10-15":
                            parliament_years.append(197576)
                        else:
                            parliament_years.append(1975)
                    else:
                        parliament_years.append(1975)
            elif str(start_year) + str(end_year)[2:] in year_list:
                try:
                    df = riksmote.loc[riksmote["parliament_year"] == int(str(start_year) + str(end_year)[2:])]
                    assert df.empty == False
                except:
                    print("wan sani de fokopu, yere -- [-1]")
                    print(int(str(start_year) + str(end_year)[2:]))
                    print(df)
                else:
                    df_ends = df["end"].unique()
                    df_starts = df["start"].unique()
                    if start < df_starts[0]:
                        if str(int(start_year)-1) + str(start_year)[2:] in year_list:
                            parliament_years.append(int(str(int(start_year)-1) + str(start_year)[2:]))
                        elif str(start_year) in year_list:
                            parliament_years.appenr(start_year)
                        else:
                            if start_year == "2000":
                                parliament_years.append(19992000)
                            else:
                                parliament_years.append(int(start_year)-1)
                    if end > df_ends[0]:
                        if str(int(end_year) + 1) + f"{int(str(end_year)[2:]) + 2:02}" in year_list:
                            parliament_years.append(int(str(int(end_year) + 1) + f"{int(str(end_year)[2:]) + 2:02}"))
                        elif end_year == "1999":
                            parliament_years.append(200001)
                        elif end_year == "1998":
                            parliament_years.append(19992000)
                        else:
                            print(int(str(int(end_year) + 1) + f"{int(str(end_year)[2:]) + 2:02}"), "isn't in year list, but maybe it should be")
                parliament_years.append(int(str(start_year) + str(end_year)[2:]))
            else:
                parliament_years.append(int(start_year))
                parliament_years.append(int(end_year))

        # if there's range
        else:
            y_range = list(range(int(start_year[:4]), int(end_year[:4])+1))
            if debug:
                print(y_range)
            if y_range[0] in year_list and y_range[-1] in year_list:
                [parliament_years.append(int(_)) for _ in y_range]
            else:
                for ix, y in enumerate(y_range):
                    #to_append = []
                    if y > 2023:
                        continue
                    if str(y) in year_list and str(y) != "1975":
                        parliament_years.append(int(y))
                    else:
                        if str(y) == "1999":
                            to_append = ["19992000"]
                            y = 19992000
                        elif str(y) == "1975":
                            if ix == 0:
                                if end_year and end_year > "1975":
                                    y, to_append = 197576, [1975, 197576]
                                elif end_year and str(end_year) == "1975":
                                    if start_p == "day":
                                        if start > "1975-10-15":
                                            y, to_append = 197576, [197576]
                                        else:
                                            y, to_append = 1975, [1975, 197576]
                                else:
                                    y, to_append = 1975, [1975]
                            elif ix == len(y_range)-1:
                                if end_p == "day":
                                    if end >= "1975-10-15":
                                        y, to_append = 197576, [1975, 197576]
                                    else:
                                        y, to_append = 1975, [1975]
                                else:
                                    y, to_append = 1975, [1975]
                            else:
                                y, to_append = 197576, [1975, 197576]
                        else:
                            to_append = [int(str(y) + f"{int(str(y)[2:])+1:02}")]

                        if debug:
                            print(ix, y, to_append)
                        for a in to_append:
                            if ix == 0:
                                try:
                                    df = riksmote.loc[riksmote["parliament_year"] == int(a)]
                                    assert df.empty == False
                                except:
                                    print("wan sani de fokopu, yere -- [0]")
                                    print(a)
                                    print(df)
                                    continue
                                else:
                                    if start_p == "day":
                                        df_ends =  df["end"].unique()
                                        if debug:
                                            print(start, df_ends, start >= df_ends[0])
                                        if start <= df_ends[0]:
                                            parliament_years.append(a)
                                        else:
                                            x = f"{int(str(start_year)[2:])-1:02}"
                                            if x == "-1":
                                                xx = 19992000
                                            else:
                                                xx = int(str(start_year)[:2] + x + f"{int(str(start_year)[2:]):02}")

                                            parliament_years.append(xx)
                                    else:
                                        parliament_years.append(a)

                            elif ix == len(y_range)-1:
                                try:
                                    df = riksmote.loc[riksmote["parliament_year"] == int(a)]
                                    assert df.empty == False
                                except:
                                    print("wan sani de fokopu, yere -- [-1]")
                                    print(a)
                                    print(df)
                                    continue
                                else:
                                    if end_p == "day":
                                        df_starts =  df["start"].unique()
                                        if debug:
                                            print(end, df_starts, end <= df_starts[0])
                                        if end <= df_starts[0]:
                                            parliament_years.append(a)

                                        else:
                                            x = f"{int(str(end_year)[2:])-1:02}"
                                            if x == "-1":
                                                xx = 19992000
                                            else:
                                                xx = int(str(end_year)[:2] + x + f"{int(str(end_year)[2:]):02}")
                                            parliament_years.append(xx)
                                    else:
                                        parliament_years.append(a)
                            else:
                                parliament_years.append(a)
    return list(set(parliament_years))



def yearize_date(date, riksmote):
    """
    takes a date and a dataframe of riksdagen-persons/data/riksdag-year.csv
    """
    parliament_years = riksmote["parliament_year"].unique()
    parliament_years = [str(_) for _ in parliament_years]
    if len(date) == 10:
        date_p = "day"
    elif len(date) == 4:
        date_p = "year"
    else:
        print(f"  ERROR!!!\nI can't handle dates of len {len(date)}")
        sys.exit()
    year = date[:4]

    return _get_parliament_years(date, None, year, None, date_p, None, parliament_years, riksmote, None)[0]




def yearize_mandates(debug_id=None, metadata_folder="data"):
    """
    return a dataframe of mandate periods, lengthened for each parliament year

    from
        mp   start        end
        A    2000-10-01   2002-09-30
    to
        mp   start       end           parliament_year
        A    2000-10-01  2001-09-30    200001
        A    2001-10-01  2002-09-30    200201

    """
    print("Yearizing mandates... this will take a moment.")
    riksmote = pd.read_csv(f"{metadata_folder}/riksdag-year.csv")
    mep = pd.read_csv(f"{metadata_folder}/member_of_parliament.csv")
    minister = pd.read_csv(f"{metadata_folder}/minister.csv")
    speaker = pd.read_csv(f"{metadata_folder}/speaker.csv")
    government = pd.read_csv(f"{metadata_folder}/government.csv")
    parliament_years = riksmote["parliament_year"].unique()
    parliament_years = [str(_) for _ in parliament_years]
    rows = []
    slen = []
    elen = []
    cols = ["person_id", "start", "end", "role", "parliament_year"]
    for k, df in {"member_of_parliament":mep, "minister": minister, "speaker": speaker}.items():
        if k == "minister":
            df = impute_minister_date(impute_date(df, metadata_folder), government)
        if k == "speaker":
            df = impute_speaker_date(impute_date(df, metadata_folder))
        for i, r in tqdm(df.iterrows(), total=len(df)):

            if debug_id and r['person_id'] == debug_id:
                print(r)

            start = r["start"]
            end = r["end"]
            start_p = None
            end_p = None
            start_y = None
            end_y = None

            if pd.notnull(start):
                if "timestamps.Timestamp" in str(type(start)):
                    start_p = "year"
                    start_y = start.strftime('%Y')
                    start = start.strftime('%Y-%m-%d')
                else:
                    if len(str(start)) not in slen:
                        slen.append(len(str(start)))
                    if len(str(start)) == 10:
                        start_p = "day"
                        start_y = str(start)[:4]
                    elif len(str(start)) == 4:
                        start_p = "year"
                        start_y = str(start)
                    elif len(str(start)) == 6:
                        start_p = "year"
                        start_y = str(start)[:4]
                    elif len(str(start)) == 7:
                        print(r)
                    elif len(str(start)) > 10:
                        start = start[:10]
                        start_y = start[:4]
                        start_y = "year"
                    else:
                        print(r, len(start))

            if pd.notnull(end):
                if "timestamps.Timestamp" in str(type(end)):
                    end_p = "year"
                    end_y = end.strftime('%Y')
                    end = end.strftime('%Y-%m-%d')
                else:
                    if len(str(end)) not in elen:
                        elen.append(len(str(end)))
                    if len(str(end)) == 10:
                        end_p = "day"
                        end_y = str(end)[:4]
                    elif len(str(end)) == 4:
                        end_p = "year"
                        end_y = str(end)
                    elif len(str(end)) == 6:
                        end_p = "year"
                        end_y = str(end)[:4]
                    elif len(str(end)) == 7:
                        print(r)
                    elif len(str(end)) > 10:
                        end = end[:10]
                        end_y = end[:4]
                        end_y = "year"
                    else:
                        print(r, len(end))

            if start_y and 1866 < int(start_y[:4]) < 2025:
                if debug_id and debug_id == r["person_id"]:
                    pys = _get_parliament_years(start, end, start_y, end_y, start_p, end_p, parliament_years, riksmote, True)
                else:
                    pys = _get_parliament_years(start, end, start_y, end_y, start_p, end_p, parliament_years, riksmote, None)
                if debug_id and r['person_id'] == debug_id:
                    print(start_y, end_y, pys, start_p, end_p)
                for ix, py in enumerate(pys):
                    if int(py) < 202324:
                        try:
                            py_start = riksmote.loc[riksmote["parliament_year"] == int(py), "start"].unique()[0]
                        except:
                            print("start")
                            print(py, type(py), pys, start, end, start_y, end_y)
                        try:
                            py_end = riksmote.loc[riksmote["parliament_year"] == int(py), "end"].unique()[-1]
                        except:
                            print("end")
                            print(py, pys, start, end, start_y, end_y)

                        if start_p == "year":
                            start = py_start
                        if end_p == "year":
                            end = py_end

                        if len(pys) == 1:
                            rows.append([r["person_id"], start, end, r["role"], int(py)])
                        else:
                            if ix == 0:
                                rows.append([r["person_id"], start, py_end, r["role"], int(py)])
                            elif ix == len(pys)-1:
                                rows.append([r["person_id"], py_start, end, r["role"], int(py)])
                            else:
                                rows.append([r["person_id"], py_start, py_end, r["role"], int(py)])

            else:
                pass
    df = pd.DataFrame(rows, columns=cols)
    df.sort_values(by="person_id", inplace=True)
    return df



def test_yearize_mandates():
    m = yearize_mandates(debug_id="i-XvpBFhfR9SDB1i9KLCmqX2")
    #print(m.head(50))
    #print(m.sample(50).head(50))
    #print(sorted(m["parliament_year"].unique()))
    m.to_csv("_scripts/chairs/yearized_mandates.csv", index=False)


def test_yearize_date():
    riksmote = pd.read_csv("data/riksdag-year.csv")
    print(yearize_date("1982-04-13", riksmote))
    print(yearize_date("1982-12-05", riksmote))
    print(yearize_date("1882-04-13", riksmote))
    print(yearize_date("1882-12-05", riksmote))
    print(yearize_date("1932-04-13", riksmote))
    print(yearize_date("1942-12-05", riksmote))


def main():
    test_yearize_mandates()
    #test_yearize_date()



if __name__ == '__main__':
    main()



"""
General utilities related to curation of Riksdagen motions.
"""
from pyriksdagen.utils import parse_tei



def fetch_template(loc="input/motions/mot-template.xml" ):
    root, ns = parse_tei(loc)
    return root, ns



"""
Dry Argparse helper.

"""
from datetime import datetime
from glob import glob
from pyriksdagen.utils import (
    get_data_location,
    protocol_iterators,
    corpus_iterator,
)
import argparse
import inspect
import sys




def populate_common_arguments(parser):
    """
    add arguments common to all doctypes to a parser object
        - start
        - end
        - year
        - from-list
        - specific-files
        - data-folder
    Args:
        parser: argparse Parser

    Returns:
        parser: argparse Parser
    """
    parser.add_argument("-s", "--start",
                        type=int,
                        default=None,
                        help="start year. If -s is set, -e must also be set. If -s and -e are explicitly set, all protocols in that range are processed. If -s and -e are unset and neither -p nor -l are set, the script will process all records from 1867 until the current year. Priority == 4")
    parser.add_argument("-e", "--end",
                        type=int,
                        default=None,
                        help="end year. If -e is set, -s must also be set. If -s and -e are explicitly set, all protocols in that range are processed. If -s and -e are unset and neither -p nor -l are set, the script will process all records from 1867 until the current year. Priority == 4")
    parser.add_argument("-y", "--parliament-year",
                        type=int,
                        default=None,
                        nargs='*',
                        help="Parliament year, e.g. 1971 or 198384. Priority == 3")
    parser.add_argument("-l", "--from-list",
                        type=str,
                        default=None,
                        help="operate on a list of records from a file. Set the path from ./ -- this option doesn't cooperate with `-R`. Priority == 2")
    parser.add_argument("-f", "--specific-files",
                        type=str,
                        default=None,
                        nargs='*',
                        help="operate on a specific file or list of files. Set the path from ./ -- this option doesn't cooperate with `-R`. User is responsible to ensure the file is the correct doctype. Priority == 1")
    parser.add_argument("-L", "--data-folder",
                        type=str,
                        default=None,
                        help="(optional) Path to folder containing the data, defaults to environment variable according to the document type or `data/` if no suitable variable is found.")
    return parser




def rename_file_list(args):
    """
    Renames the specific_files arg to a doctype-relevant name

    Args:
        args: argparse Namespace
        doctype (str): document type

    Returns:
        args
    """
    d = vars(args)
    d[d["doctype"]] = d.pop("specific_files")
    return argparse.Namespace(**d)




def common_args(args):
    """
    common preprocessing of arguments

    Args:
        args: argparse Argument namespace

    Returns:
        args
    """
    if (args.start is None or args.end is None) and args.start != args.end:
        raise ValueError("Set -s and -e or neither.")

    if args.data_folder is None:
        args.data_folder = get_data_location(args.doctype)

    if args.specific_files is not None and len(args.specific_files) != 0:
        pass
    elif args.from_list is not None:
        with open(args.from_list, 'r') as inf:
            lines = inf.readlines()
            args.specific_files = [_.strip() for _ in lines if _.strip() != '']
    elif args.parliament_year is not None and len(args.parliament_year) != 0:
        args.specific_files = []
        for py in args.parliament_year:
            args.specific_files.extend(glob(f"{args.data_folder}/{py}/*.xml"))
    else:
        args.specific_files = sorted(list(corpus_iterator(args.doctype, corpus_root=args.data_folder, start=args.start, end=args.end)))
    if args.specific_files is not None and len(list(args.specific_files)) != 0:
        args.specific_files = sorted(list(args.specific_files))
    return rename_file_list(args)




def record_parser(parser):
    """
    Take an argparse ArgumentParser object and populate standard arguments for working with riksdagen records.

    Args:
        parser: parser

    Returns:
        parser
    """
    parser = populate_common_arguments(parser)
    # leaving the records-specific function in place, in case we need records-specific args
    return parser




def record_args(args):
    """
    Takes an argparse namespace object for working with riksdagen records and imputes standard stuff

    Args:
        args: args

    Returns:
        args
    """
    args = common_args(args)
    # leaving the records-specific function in place, in case we need records-specific actions
    return args




def motion_parser(parser):
    """
    Take an argparse ArgumentParser object and populate standard arguments for working with riksdagen motions.

    Args:
        parser: parser

    Returns:
        parser
    """
    parser = populate_common_arguments(parser)
    # leaving the motions-specific function in place, in case we need motions-specific args
    return parser




def motion_args(args):
    """
    Takes an argparse namespace object for working with riksdagen records and imputes standard stuff

    Args:
        args: args
    Returns:
        args
    """
    args = common_args(args)
    # leaving the records-specific function in place, in case we need records-specific actions
    return args




def interpellation_parser(parser):
    """
    Take an argparse ArgumentParser object and populate standard arguments for working with riksdagen interpellations.

    Args:
        parser: parser

    Returns:
        parser
    """
    parser = populate_common_arguments(parser)
    # leaving the interpellations-specific function in place, in case we need interpellations-specific args
    return parser




def interpellation_args(args):
    """
    Takes an argparse namespace object for working with riksdagen interpellations and imputes standard stuff

    Args:
        args: args

    Returns:
        args
    """
    args = common_args(args)
    # leaving the interpellations-specific function in place, in case we need interpellations-specific actions
    return args




def fetch_parser(doctype, docstring=None):
    """
    Fetch an argparse argument parser based on the doctype.

    Args:
        doctype (str): doctype, one listed in D
        docstring (str): string describing scripts for which the parser is called.

    Returns:
        argparse Parser object
    """
    D = {
            "records": record_parser,
            "motions": motion_parser,
            "interpellations": interpellation_parser,
        }
    parser = argparse.ArgumentParser(description=docstring)
    parser.add_argument("--doctype", default=doctype, help=argparse.SUPPRESS)
    return D[doctype](parser)




def impute_args(args):
    """
    Impute args based on the doctype (args.doctype).

    Args:
        args: argparse parsed args namespace

    Returns:
        argparse parsed args namespace
    """
    D = {
            "records": record_args,
            "motions": motion_args,
            "interpellations": interpellation_args,
        }
    return D[args.doctype](args)




# test
if __name__ == '__main__':
    args = fetch_parser("motions").parse_args()
    print(impute_args(args))
